(* ald_pm.v *)
(* Gives a refinement of ALD to PM *)
(* Also includes proofs that the specifications of ALD's primitives are implemented by functions in the PM *)
(* Also shows that the ALD spec of ptinit matches up with PD's spec of ptinit *)
(* TODO: BUGS!!!! *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.
Require Import mem.

Require Import mem_repr.
Require Import pm.
Require Import ald.
Require Import ald_lib.
Require Import mem_sig_ald.


Require Import pe.
Opaque PE_State.PMM_data_loc.

Module ALD_PM <: Mem_Repr_T ALD_State PM_State.
Import mem_sig_ald.

Definition Z_to_bool (z : Z) := negb (Zeq_bool z 0).

Definition Repr_Mem (ald : ALD_State.t) (pm : PM_State.t) :=
   (forall l, ValidPg (Pg l) = true -> 
                (ALD_State.alloctbl ald (Pg l) = true) -> 
                ALD_State.data ald l = pm l) /\
   (forall l, ValidPA l = true -> PE_State.PMMDom l = false ->
            ALD_State.alloctbl ald (Pg l) = true ->
            pm (PE_State.PMM_data_loc + (Pg l) * 8)%Z = 1%Z) /\
   (forall l, ValidPA l = true -> PE_State.PMMDom l = false ->
            ALD_State.alloctbl ald (Pg l) = false ->
            pm (PE_State.PMM_data_loc + (Pg l) * 8)%Z = 0%Z) /\
   (forall l, ValidPA l = true -> PE_State.PMMDom l = true ->ALD_State.alloctbl ald (Pg l) = false) /\
   (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> pm (PE_State.PMM_data_loc + (Pg l) * 8)%Z = 1%Z)
.

Lemma Repr_getMemCheck : forall (m1 : ALD_State.t) (m2 : PM_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, ALD_State.getMemCheck l m1 = true -> PM_State.getMemCheck l m2 = true.
intros.
red in H; repeat autoclean.
unfold ALD_State.getMemCheck, PM_State.getMemCheck in *.
unfold ALD_State.ValidAddr in *.
repeat autoclean; autosolve.
Qed.

Lemma Repr_getMem : forall (m1 : ALD_State.t) (m2 : PM_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v, ALD_State.getMemCheck l m1 = true -> 
              ALD_State.getMem l m1 = v -> PM_State.getMem l m2 = v.
intros.
red in H; repeat autoclean.
unfold ALD_State.getMemCheck in *.
unfold ALD_State.ValidAddr in *.
repeat autoclean; autosolve.
unfold PM_State.getMem, ALD_State.getMem.
rewrite H; autosolve.
apply ValidPA_ValidPg; autosolve.
Qed.

Lemma Repr_setMemCheck : forall (m1 : ALD_State.t) (m2 : PM_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, ALD_State.setMemCheck l m1 = true -> PM_State.setMemCheck l m2 = true.
intros.
red in H; repeat autoclean.
unfold ALD_State.setMemCheck, PM_State.setMemCheck in *.
unfold ALD_State.ValidAddr in *.
repeat autoclean; autosolve.
Qed.

Lemma Repr_setMem : forall (m1 : ALD_State.t) (m2 : PM_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v , ALD_State.setMemCheck l m1 = true -> 
             Repr_Mem (ALD_State.setMem l v m1) (PM_State.setMem l v m2).
unfold ALD_State.setMemCheck, ALD_State.setMem, PM_State.setMem.
unfold ALD_State.ValidAddr.
intros.
red in H; repeat autoclean.
red; simpl.
simpl in *; repeat autoclean.
repeat split; autosolve.

intros.
rewrite H; autosolve.

intros.
rewrite H1; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H3 in H0; autosolve.
apply PE_State.PMMDom_Loc; auto.
apply ValidPA_ValidPg; auto.

intros.
rewrite H2; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H3 in H0; autosolve.
apply PE_State.PMMDom_Loc; auto.
apply ValidPA_ValidPg; auto.


intros.
rewrite H4; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H3 in H0; autosolve.
apply PE_State.PMMDom_Loc; auto.
apply ValidPA_ValidPg; auto.
Qed.
End ALD_PM.

Require Import c.

Module C_ALD_PM := C_Refine ALD_State PM_State ALD_PM C_ALD C_PM.

Import ALD_PM.
Require Import ald_lib.
Import ALD_Lib.
Require Import mem_sig_ald.
Import mem_sig_ald.
Require Import ald.

Opaque PE_State.PTROOT PTROOT PE_State.PMM_data_loc.

Lemma exists_alloc : forall pg newpagesrc s, exists s', 
   allocpg pg s s' /\  (forall l, Pg l = pg -> ALD_State.getMem l (C_ALD.State.memory s') = newpagesrc l).
intros.
unfold ALD_Lib.allocpg.
destruct s; simpl in *.
destruct memory; simpl in *.
exists 
{| C_ALD.State.stack := (C_ALD.State.Stack.pop +++ C_ALD.State.Stack.setRet pg) stack;
    C_ALD.State.memory := 
         {| ALD_State.data := (fun l => if Zeq_bool (Pg l) pg then newpagesrc l else data l);
             ALD_State.alloctbl := (fun x : Z => if Zeq_bool x pg then true else alloctbl x)
          |}
|}; simpl.
repeat split; auto.
intros; auto.
Zeq_pull'; Zeq_norm; repeat autoclean; simpl in *; autosolve.
intros; auto.
Zeq_pull'; Zeq_norm; repeat autoclean; simpl in *; autosolve.
Qed.


(*

Lemma alloc_weak1 : ActionGTE (C_ALD_PM.RF.reprAction C_ALD_PM.CRepr ALD_Lib.mem_alloc_spec)
  mem_alloc_spec.
split.

(* pre *)
intros.
unfold mem_alloc_spec; simpl.
unfold ALD_Lib.mem_alloc_spec in *; simpl in *.
repeat autoclean.
unfold PMM_OK, PMM_valid.
repeat split; autosolve.
red in H; repeat autoclean.
unfold C_ALD.State.StackCheck, C_PM.State.StackCheck in *.
rewrite <- H; autosolve.
red in H; repeat autoclean.
unfold C_ALD.State.StackCheck, C_PM.State.StackCheck in *.
rewrite <- H; autosolve.

(* rel *)
intros.
unfold mem_alloc_spec in *; simpl in *.
repeat autoclean.
inversion_clear H0; repeat autoclean.

(* x0 is allocated *)
destruct (exists_alloc x0 (C_PM.State.memory s) bs) as [bs'].
exists bs'.
unfold allocpg in *; repeat autoclean.
split.

destruct bs; destruct bs'; simpl in *; repeat autoclean.
red; simpl.
split; autosolve.
red in H3; simpl in *; repeat autoclean; autosolve.
red; simpl.
repeat split; autosolve.
intros.
rewrite H10.
unfold allocupd in *; simpl.
rewrite getPageInfo_setPageInfo.
red in H3; simpl in *; repeat autoclean; autosolve.
red in H13; repeat autoclean.
rewrite H3; autosolve.
intros.
rewrite H10.
unfold allocupd in *; simpl.
red in H3; simpl in *; repeat autoclean; autosolve.
red in H13; repeat autoclean.
rewrite H13; autosolve.
unfold C_PM.State.MemGet in H6.
Zeq_pull'; Zeq_norm; try subst; simpl in *; autosolve.
rewrite H14 in H6; autosolve.
intros.
rewrite getPageInfo_setPageInfo.
red in H3; simpl in *; repeat autoclean; autosolve.
red in H13; repeat autoclean.
rewrite H14; autosolve.
makenew' (Zeq_bool (Pg l) x0); autosolve.
intros.
red in H3; simpl in *; repeat autoclean; autosolve.
red in H13; simpl in *; repeat autoclean.
unfold setPageInfo.
unfold PM_State.getMem, PM_State.setMem.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H13 in H12.
destruct (Zeq_dec (Pg l) x0).
rewrite H9; auto.
unfold setPageInfo.
simpl in *.
assert (PMMDom x0 = false).
makenew' (PMM_page x0); autosolve.
unfold C_PM.State.MemGet in H6.
rewrite H14 in H6; autosolve.
destruct (Zeq_dec (Pg l) x0).
subst.
rewrite H9; autosolve.
rewrite getMem_setPageInfo; autosolve.
rewrite H10 in *; clear H10.
unfold allocupd in H12.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H7 in *; autosolve.
rewrite getMem_setPageInfo; autosolve.
apply H15; autosolve.
unfold ValidPg; Zeq_pull'; Zeq_norm; autosolve.
makenew' (PMM_page (Pg l)); autosolve.
rewrite H13 in H12; autosolve.

(**************)
left.
exists x0.
repeat split; autosolve.
red in H3; repeat autoclean; autosolve.
red in H12; repeat autoclean; autosolve.
rewrite H12; autosolve.
unfold ValidPg; Zeq_pull'; Zeq_norm; autosolve.
makenew' (PMM_page x0); autosolve.
unfold C_PM.State.MemGet in H6.
rewrite H14 in H6; autosolve.

(* failure case *)
exists (allocfail bs).
unfold allocfail in *; simpl in *.
red in H3; repeat autoclean.
unfold C_ALD.State.StackAct, C_PM.State.StackAct; simpl in *.
rewrite H0.
red in H3; repeat autoclean.
repeat split; try red; simpl; autosolve.
right; autosolve.
Qed.

Lemma free_weak1 : ActionGTE (C_ALD_PM.RF.reprAction C_ALD_PM.CRepr ALD_Lib.mem_free_spec) 
                                    mem_free_spec.
split.

(* pre *)
intros.
unfold mem_free_spec, ALD_Lib.mem_free_spec in *; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_ALD.State.StackGet, C_ALD.State.StackCheck in *.
unfold C_PM.State.StackGet, C_PM.State.StackCheck.
rewrite H in *.
unfold ValidPg in H1; Zeq_pull'; Zeq_norm; simpl in *; autosolve.
repeat split; simpl in *; autosolve.
makenew (nth 0 (C_PM.State.Stack.getArgs (C_PM.State.stack s)) 0%Z).
autosolve.
unfold C_PM.State.MemGet.
red in H4; repeat autoclean; autosolve.

rewrite <- H4; autosolve.
makenew (nth 0 (C_PM.State.Stack.getArgs (C_PM.State.stack s)) 0%Z).
unfold ValidPg; Zeq_pull'; Zeq_norm; simpl; autosolve.
makenew' (PMM_page (nth 0 (C_ALD.State.Stack.getArgs (C_PM.State.stack s)) 0%Z)); autosolve.
rewrite H7 in H2; autosolve.

(* rel *)
intros.
unfold mem_free_spec, ALD_Lib.mem_free_spec in *; simpl in *; repeat autoclean.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold C_PM.State.MemGet in *.

red in H2; repeat autoclean.
unfold C_ALD.State.StackGet, C_ALD.State.StackCheck in *.
unfold C_PM.State.StackGet, C_PM.State.StackCheck in *.
eexists; split; autosolve.
makenew' (getPageInfo
          (nth 0 (C_PM.State.Stack.getArgs (C_PM.State.stack s)) 0%Z)
          (C_PM.State.memory s)); simpl in *; autosolve.
subst; simpl.
red; simpl.
rewrite H2.
split; autosolve.
red in H10; repeat autoclean.
red; repeat split; autosolve.
unfold allocupd; simpl.
intros.
rewrite getPageInfo_setPageInfo.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H0; autosolve.

unfold allocupd; simpl.
intros.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H10; autosolve.

intros.
rewrite getPageInfo_setPageInfo.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H10 in H5; autosolve.
rewrite H2; autosolve.

apply H11; autosolve.

simpl.
unfold allocupd.
intros.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite getMem_setPageInfo; autosolve.
rewrite <- H13; autosolve.
rewrite H2 in *.
unfold C_ALD.State.Stack.getArgsCheck in H3.
makenew' (C_PM.State.stack s); simpl in *; autosolve.
destruct f; simpl in *; autosolve.
unfold ValidPg; Zeq_pull'; Zeq_norm; autosolve.
makenew' (PMM_page (Pg l)); autosolve.
rewrite H10 in H15; autosolve.
Qed.
*)


Require Import mem_sig.
Require Import ale_pe.
Require Import ale.


Lemma exists_mem : forall s, 
(forall l : Z,
     ValidPA l = true ->
     (s (pe.PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z \/
      s (pe.PE_State.PMM_data_loc + Pg l * 8)%Z = 0%Z)
     ) ->  
(forall l : Z,
     ValidPA l = true ->
     PE_State.PMMDom l = true ->
     s (pe.PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) ->
(forall l : Z,
     ValidPA l = true ->
     ALE_State.PtDom l = true ->
     s (pe.PE_State.PMM_data_loc + Pg l * 8)%Z = 1%Z) ->
exists bs, Repr_Mem bs s /\
(forall l : Z,
ValidPA l = true ->
ALE_State.PtDom l = true -> ALD_State.alloctbl bs (Pg l) = true).

intros.
exists {| ALD_State.data := s; ALD_State.alloctbl := 
       fun pg => if PE_State.PMMDom (pg * PGSIZE + Off 0)%Z then false else
                      Zeq_bool (s (PE_State.PMM_data_loc + pg * 8)%Z) 1%Z
 |}.
split.
red; simpl.
repeat split; autosolve.

intros.
makenew' (PE_State.PMMDom (Pg l * 4096 + Off 0)); autosolve.

intros.
rewrite <- PE_State.PMMDom_zero in *.
rewrite H3 in H4.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
destruct (H l); autosolve.

intros.
rewrite H0; autosolve.
rewrite <- PE_State.PMMDom_zero.
rewrite H3; auto.

intros; simpl.
rewrite <- PE_State.PMMDom_zero.
rewrite H1; auto.
rewrite Zeq_bool_refl.
makenew' (PE_State.PMMDom l); auto.
elimtype False.
apply ale_pe.PMM_PT_separate' with l; auto.
Qed.


Lemma pmap_init_weak : ActionGTE mem_sig_ald.PM_pmap_init_spec
            (C_ALD_PM.RF.reprAction C_ALD_PM.CRepr ALD_Lib.ALD_pmap_init_spec).
split.

unfold mem_sig_ald.PM_pmap_init_spec, ALD_pmap_init_spec; simpl; intros.
unfold C_PM.State.StackCheck, C_ALD.State.StackCheck in *; repeat autoclean.
unfold C_ALD_PM.CRepr.
unfold Repr_Mem.

destruct (exists_mem (C_PM.State.memory s)) as [bmem]; autosolve.
destruct H3.
destruct s; simpl in *.
exists {| C_ALD.State.stack := stack; C_ALD.State.memory := bmem |}; simpl in *.
red in H3; repeat autoclean.
repeat split; autosolve.

(* rel *)
intros.
unfold mem_sig_ald.PM_pmap_init_spec, ALD_pmap_init_spec in *; simpl in *; intros.
repeat autoclean.
destruct s; simpl in *.
destruct (exists_mem memory) as [bmem]; autosolve.
destruct H4.
specialize H0 with
 {| C_ALD.State.stack := stack; C_ALD.State.memory := bmem |}; simpl in *; repeat autoclean.
assert (C_ALD_PM.CRepr
       {| C_ALD.State.stack := stack; C_ALD.State.memory := bmem |}
       {| C_PM.State.stack := stack; C_PM.State.memory := memory |}).
red; simpl; split; auto.
specialize (H0 H6).
clear H6.
destruct H0; repeat autoclean.
unfold C_ALD.State.StackCheck, C_PM.State.StackCheck in *; simpl in *.
split; auto.
(*****)
repeat split.

intros.
red in H0; repeat autoclean.
red in H12; repeat autoclean.
rewrite <- H12; auto.
apply ValidPA_ValidPg.
apply ALE_State.PtDom_Valid; auto.
apply ValidPg_ValidVPg; auto.
rewrite H9; auto.
apply H5; auto.
apply ALE_State.PtDom_Valid; auto.
apply ValidPg_ValidVPg; auto.
apply ALE_State.PtDom_PTROOT; auto.
apply ValidPg_ValidVPg; auto.
apply ValidPA_ValidPg.
apply ALE_State.PtDom_Valid; auto.
apply ValidPg_ValidVPg; auto.

intros.
red in H0; repeat autoclean.
red in H13; repeat autoclean.
rewrite <- H13; auto.
apply ValidPA_ValidPg.
apply ALE_State.PtDom_Valid; auto.
rewrite H9; auto.
apply H5; auto.
apply ALE_State.PtDom_Valid; auto.
apply ALE_State.PtDom_PTROOT; auto.
apply ValidPA_ValidPg.
apply ALE_State.PtDom_Valid; auto.



intros.
red in H0; repeat autoclean.
red in H12; repeat autoclean.
rewrite <- Pg_simpl with pg 0%Z.
destruct (bool_dec (PE_State.PMMDom (pg * PGSIZE + Off 0)) true).
rewrite H16; auto.
rewrite H2; auto.
apply ValidPg_ValidPA_zero; auto.
apply ValidPg_ValidPA_zero; auto.
assert (PE_State.PMMDom (pg * PGSIZE + Off 0) = false).
makenew' ((PE_State.PMMDom (pg * PGSIZE + Off 0))); autosolve.
elimtype False; apply n; auto.
clear n.

destruct (bool_dec (ALD_State.alloctbl (C_ALD.State.memory x) (Pg (pg * 4096 + Off 0)))  true).
rewrite H13; autosolve.
red in H4; repeat autoclean.
rewrite H18; autosolve.
apply ValidPg_ValidPA_zero; auto.
rewrite <- H9; autosolve.
rewrite Pg_simpl; auto.
apply ValidPg_ValidPA_zero; auto.

assert ((ALD_State.alloctbl (C_ALD.State.memory x) (Pg (pg * 4096 + Off 0))) = false).
red in n.
makenew' ((ALD_State.alloctbl (C_ALD.State.memory x) (Pg (pg * 4096 + Off 0)))); autosolve.
elimtype False; apply n; auto.
clear n.
rewrite H14; autosolve.
red in H4;repeat autoclean.
rewrite H20; autosolve.
apply ValidPg_ValidPA_zero; auto.
rewrite <- H9; autosolve.
rewrite Pg_simpl; auto.
apply ValidPg_ValidPA_zero; auto.


red in H0; repeat autoclean.
rewrite <- H0.
rewrite H10; auto.

Qed.



(******************************************************)
Require Import label.

Definition ref := C_ALD_PM.EqRef.

Lemma WF_ALD_pm : forall CODE PSI,
         C_ALD.SCAP.WFCode C_ALD.State.expr_to_Prop C_ALD.Mc_Machine 
                               C_ALD.SCAP.SpecHeap.SpecHeap_empty CODE PSI ->

      C_PM.SCAP.WFCode C_PM.State.expr_to_Prop C_PM.Mc_Machine 
                C_PM.SCAP.SpecHeap.SpecHeap_empty 
                (C_ALD_PM.CMake CODE)
                (C_ALD_PM.PsiMake PSI).
intros.
set (x := ref C_ALD.SCAP.SpecHeap.SpecHeap_empty CODE PSI).
apply x; auto.
Qed.
