(* ale.v *)
(* Defines the ALE memory model *)
(* TODO: Currently includes the TLB state which is not reflected in the actual semantics *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem.

Module ALE_State <: Mem_T.

Definition PTROOT := PTROOT.


Record State := { data : Z -> Z; allocdata : Z -> bool; tlb : bool }.
Definition t := State.


Definition PtDom (l : Z) := Zle_bool PTROOT l && Zlt_bool l (PTROOT + VPAGES * 8).
Lemma PtDom_alloc : forall l, PtDom l = true -> (100 <= Pg l < 512)%Z.
intros.
unfold PtDom in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold PTROOT, common.PTROOT in *.
unfold Pg.
split.
apply Zle_mult_div; try omega.
apply Zlt_mult_div_1; try omega.
Qed.

Lemma PtDom_zero : forall l, PtDom l = PtDom (Pg l * PGSIZE + Off 0).
intros.
rewrite Off0.
unfold PtDom, Pg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve; elimtype False;
unfold PTROOT, common.PTROOT in *; simpl in *.

replace (1507328)%Z with (368 * 4096)%Z in H3; try omega.
assert (l / 4096 >= 368)%Z; try omega.
clear H3.
assert (l >= 368 * 4096)%Z.
apply Zge_div_mult; autosolve.
omega.

replace (1507328)%Z with (368 * 4096)%Z in *; try omega.
assert (l / 4096 < 368)%Z; try omega.
clear H3.
assert (l / 4096 >= 368)%Z.
apply Zge_mult_div_1; autosolve.
omega.

replace (1441792)%Z with (352 * 4096)%Z in *; try omega.
assert (352 > l / 4096)%Z; try omega.
clear H1.
assert (352 <= l / 4096)%Z.
apply Zle_mult_div; autosolve.
omega.

replace (1441792)%Z with (352 * 4096)%Z in *; try omega.
assert (352 <= l / 4096)%Z; try omega.
clear H1.
assert (352 * 4096 <= l)%Z; try omega.
apply Zle_div_mult; try omega.
assert (0 < l / 4096)%Z; try omega.
apply Zgt_div_0 with PGSIZE; try omega.
Qed.

Opaque PTROOT.


Definition trans (l : Z) : SGet Z (State := State) :=
fun s => let ppg  := s.(data) (PTROOT + (Pg l) * 8) in
               (ppg * PGSIZE + Off l)%Z.

Definition transCheck (l : Z) : SCheck (State := State) :=
fun s =>  ValidVA l && s.(tlb) &&
               if isPhys l then Zeq_bool (trans l s) l
               else negb (PtDom (trans l s)).

Definition DM : State -> Prop  := 
   fun s => forall l, ValidPA l = true -> trans l s = l. 

Definition paCheck (l : Z) : SCheck (State := State) :=
fun s => s.(allocdata) (Pg l) && ValidPA l.

Definition vaCheck (l :Z) : SCheck (State := State) := 
fun s => transCheck l s && paCheck (trans l s) s.

Definition getMemCheck (l:Z) : SCheck (State := State) := vaCheck l.
Definition setMemCheck (l : Z) : SCheck (State := State) := fun s => vaCheck l s (* && negb (PtDom l) *).
 
Definition getMem (l:Z) : SGet (State := State) Z :=
fun s => s.(data) (trans l s).

Definition setMem (l z : Z) :  SAct (State := State) :=
fun s => let pa := trans l s in
             {| data := fun l' => if Zeq_bool pa l' then z else s.(data) l'; 
                      allocdata := s.(allocdata); tlb := s.(tlb) |}.


Definition getTLB : SGet (State := State) bool :=
fun s => s.(tlb).

Definition setTLB (t : bool) :  SAct (State := State) :=
fun s =>  {| data := s.(data); allocdata := s.(allocdata); tlb := t |}.



(* getMem simplifications *)
Lemma setMem_unneeded : forall l z s, getMemCheck l s = true -> (getMem l s = z) -> (setMem l z s = s).
intros.
unfold getMem, setMem, ValidVA, ValidPA in *; simpl; subst.
destruct s; simpl in *.
f_equal; simpl.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.
Qed.


Lemma PtDomCalc : forall l0, ValidVA l0 = true ->
PtDom (PTROOT + Pg l0 * 8) = true.
intros.
unfold ValidVA, PtDom, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve; elimtype False.
unfold Pg in H3.
assert( 0 > l0 / 4096 * 8)%Z; try omega.
clear H3.
assert (0 > l0 / 4096)%Z; try omega.
clear H0.
assert (0 <= l0 / 4096)%Z; try omega.
apply Z_div_pos; autosolve.
Qed.

Definition PtDom_allocated (s : State) := 
      forall l, ValidPA l = true -> PtDom l = true -> s.(allocdata) (Pg l) = true.

Lemma trans_setMem_unch : forall l0 l v m1,
ValidVA l0 = true -> 
PtDom l = false ->
setMemCheck l m1 = true -> 
trans l0 (setMem l v m1) = trans l0 m1.
intros.
unfold setMemCheck, vaCheck in *; repeat autoclean.
unfold transCheck in *; repeat autoclean.
makenew' (isPhys l).

Zeq_pull'; Zeq_norm.
rewrite H3 in H2.
unfold setMem.
rewrite H3.
unfold trans; simpl.
Zeq_pull'; Zeq_norm; simpl; auto.
subst.
rewrite PtDomCalc in *; autosolve.

makenew' (PtDom (trans l m1)); autosolve.
clear H3.
unfold trans.
unfold setMem.
simpl.
Zeq_pull'; Zeq_norm; simpl; auto.
rewrite H5 in H7.
rewrite PtDomCalc in *; autosolve.
Qed.

Lemma transCheck_setMem_unch : forall l0 l v m1,
PtDom l = false ->
setMemCheck l m1 = true -> 
transCheck l0 (setMem l v m1) = transCheck l0 m1.
intros.
unfold transCheck; simpl in *.
makenew' (ValidVA l0); simpl in *; autosolve.
rewrite trans_setMem_unch; autosolve.
Qed.

Lemma getMemCheck_setMem_unch : 
forall l0 l v m1, 
  PtDom l = false -> 
  setMemCheck l m1 = true -> 
  ValidVA l0 = true ->
  getMemCheck l0 (setMem l v m1) = getMemCheck l0 m1.
intros.
unfold getMemCheck, vaCheck in *; simpl in *.
rewrite trans_setMem_unch; autosolve.
rewrite transCheck_setMem_unch; autosolve.
Qed.

Definition Collision (l l' : Z) : SCheck (State := State) := 
fun s => Zeq_bool (trans l s) (trans l' s).

Lemma getMem_setMem : 
  forall l l' z s, PtDom l' = false -> 
            setMemCheck l' s = true ->
            getMemCheck l s = true ->
            getMem l (setMem l' z s) = if Collision l l' s then z else getMem l s.
intros.
unfold getMem, Collision in *; simpl in *.
rewrite trans_setMem_unch; autosolve.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
unfold getMemCheck in H1.
unfold vaCheck in H1; repeat autoclean.
unfold transCheck in H1; repeat autoclean.
autosolve.
Qed.

Lemma PtDom_PTROOT : forall vpg, ValidVPg vpg = true -> 
  ALE_State.PtDom (ALE_State.PTROOT + vpg * 8) = true.
intros.
unfold ValidVPg, PtDom, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma PtDom_Valid : forall vpg, ValidVPg vpg = true -> ValidPA (ALE_State.PTROOT + vpg * 8) = true.
intros.
unfold ValidVPg, ValidPA, VPAGES, NPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

unfold WordAligned.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Transparent PTROOT.
unfold PTROOT, common.PTROOT in *.
replace (352 * 4096)%Z with ((352 * 512) * 8)%Z in H5; try omega.
rewrite math3 in H5; try omega.
rewrite mod_elim in H5; try omega.
elimtype False; apply H5; auto.

elimtype False.
unfold Pg in *.
unfold PTROOT, common.PTROOT in *.
rewrite math1 in H4; try omega.
assert (vpg * 8 / 4096 >= 744)%Z; try omega.
clear H4.
assert (vpg * 8 >= 744 * 4096)%Z; try omega.
apply Zge_div_mult; try omega.

elimtype False.
unfold Pg in *.
unfold PTROOT, common.PTROOT in *.
assert((0 - (352 * 4096)) > vpg * 8)%Z; try omega.

elimtype False.
unfold Pg in *.
unfold PTROOT, common.PTROOT in *.
assert((0 - (352 * 4096)) > vpg * 8)%Z; try omega.
Qed.

End ALE_State. 
