(* common.v *)
(* Contains some definitions and lemmas that are used throughout the machine definitions *)
(* TODO: could use quite a bit of clean-up *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Import sact.
Require Import actsmp.

Lemma Zeq_dec (z z' : Z) : {z = z'} + {z <> z'}.
decide equality.
decide equality.
decide equality.
Qed.

Lemma Zge_divdiv : forall a b, (b > 0)%Z -> (a >= a / b * b)%Z.
intros.
destruct (Zeq_dec a 0)%Z.
subst; simpl; autosolve.

set (x := Z_div_mod a b H).
unfold Zdiv.
makenew' (Zdiv_eucl a b).
destruct x; subst; simpl.
rewrite Zmult_comm.
assert (z0 >= 0)%Z; autosolve.
Qed.

Lemma Zgt_div_0 : forall a b, (b >0)%Z -> (a / b > 0)%Z -> (a > 0)%Z.
intros.
set (x := Z_div_mod a b H).
unfold Zdiv in H0.
makenew' (Zdiv_eucl a b).
destruct x; subst; simpl.
assert(b *z > 0)%Z; autosolve.
apply Zmult_gt_0_compat; auto.
Qed.


Lemma math6 : forall a b, (b > 0)%Z -> (0 > (a / b))%Z -> (0 > a)%Z.
intros.
destruct (Ztrichotomy 0 a).
assert (0 / b <= a / b)%Z.
apply Z_div_le; autosolve.
autosolve.

destruct H1; subst; autosolve.
Qed.

Lemma Zle_mult_div: forall a b c,  (c > 0)%Z -> (a * c <= b)%Z -> (a <= b / c)%Z.
intros.
apply Zdiv_le_lower_bound; autosolve.
Qed.

Lemma Zlt_mult_div_1 : forall a b c, (c > 0)%Z -> (a < b * c)%Z -> (a / c < b)%Z.
intros.
apply Zdiv_lt_upper_bound; autosolve.
Qed.


Lemma Zge_div_mult : forall a b c, (a >= 0)%Z -> (c > 0) %Z -> (a / c >= b)%Z -> (a >= b * c)%Z.
intros.
assert (a / c * c >= b * c)%Z; try omega.
apply Zmult_ge_compat_r; autosolve.
assert (a >= a / c * c)%Z; try omega.
apply Zge_divdiv; try omega.
Qed.

Lemma Zge_mult_div_1 : forall a b c, (c > 0)%Z -> (a >= b * c)%Z -> (a / c >= b)%Z.
intros.
assert (b <= a / c)%Z; try omega.
apply Zdiv_le_lower_bound; autosolve.
Qed.

Lemma Zle_div_mult : forall a b c, (b > 0)%Z -> (c > 0)%Z -> (a <= b / c)%Z -> (a * c <= b)%Z.
intros.
assert (a * c <= b / c * c)%Z.
apply Zmult_le_compat_r; autosolve.
assert (b / c * c <= b)%Z; autosolve.
assert (b >= b / c * c)%Z; autosolve.
apply Zge_divdiv; autosolve.
Qed.

Lemma math1: forall a b c, c <> 0%Z -> ((a * c + b) / c = a + b / c)%Z.
intros.
rewrite Z_div_plus_full_l; autosolve.
Qed.

Lemma math3: forall a b c, c <> 0%Z -> ((a * c + b) mod c = b mod c)%Z.
intros.
replace (a * c + b)%Z with (b + a * c)%Z; autosolve.
rewrite Z_mod_plus_full; autosolve.
Qed.

Lemma math2 : forall a c, (c>0)%Z -> ((a mod c) / c = 0)%Z.
intros.
apply Zdiv_small; autosolve.
apply Z_mod_lt; auto.
Qed.

Lemma math4 : forall a c, (((a mod c) mod c) = (a mod c))%Z.
exact Zmod_mod.
Qed.

Lemma math5 : forall a c, (c > 0)%Z -> (a = (a / c) * c + (a mod c))%Z.
intros.
replace (a / c * c)%Z with (c * (a / c))%Z; autosolve.
rewrite <- Z_div_mod_eq_full; autosolve.
rewrite Zmult_comm; auto.
Qed.



Lemma mod0 : forall c, (0 mod c = 0)%Z.
exact Zmod_0_l.
Qed.

Lemma mod_elim : forall a c, (((a * c) mod c) = 0)%Z.
exact Z_mod_mult.
Qed.

Lemma Zplus_zero : forall a:Z, ((a + 0) = a)%Z.
intros.
omega.
Qed.

Lemma mod_plus : forall a b c, ((a + b) mod c = ((a mod c) + (b mod c)) mod c)%Z.
exact Zplus_mod.
Qed.

Lemma mod_ex : forall a c, (c > 0)%Z -> (a mod c = 0)%Z -> exists b, (a = c * b)%Z.
intros.
exists ((a / c)%Z).
rewrite <- Z_div_exact_full_2; autosolve.
Qed.

Lemma mod_le : forall a c, (c > 0)%Z -> (0 <= a)%Z -> (0 <= a mod c)%Z.
intros.
set (x := Z_div_mod a c H).
unfold Zmod.
makenew' (Zdiv_eucl a c).
destruct x; autosolve.
Qed.



Notation PGSIZE := 4096%Z.
Definition NPAGES := 4096%Z.
Definition VPAGES := 8192%Z.
Definition MEMSIZE : Z := (NPAGES * PGSIZE)%Z.



Definition Pg (addr : Z) := Zdiv addr PGSIZE.
Definition Off(addr : Z) := Zmod addr PGSIZE.



Definition WordAligned (pa : Z) : bool := Zeq_bool (pa mod 8) 0%Z.

Definition ValidPA (pa : Z) := (Zle_bool 0 pa &&  Zlt_bool (Pg pa) NPAGES && WordAligned pa %Z).
Definition ValidVA (va : Z) := (Zle_bool 0 va &&  Zlt_bool (Pg va) VPAGES && WordAligned va %Z).
(* null pointer is invalid *)

Definition ValidPg (pg : Z) := (Zle_bool 0 pg && Zlt_bool pg NPAGES).
Definition ValidVPg (pg : Z) := (Zle_bool 0 pg && Zlt_bool pg VPAGES).

Definition isPhys (addr : Z) : bool := ValidPg (Pg addr).

Lemma Pg_simpl : forall pg l, Pg (pg * PGSIZE + Off l) = pg.
intros.
unfold Pg, Off; simpl.
rewrite math1; autosolve.
rewrite math2; autosolve.
Qed.

Lemma Off_simpl : forall pg l, Off (pg * PGSIZE + Off l) = Off l.
intros.
unfold Off; simpl.
rewrite math3; autosolve.
rewrite math4; autosolve.
Qed.

Lemma Pg_Off_eq : forall l l', Pg l = Pg l' -> Off l = Off l' -> l = l'.
intros.
unfold Pg, Off in *.
rewrite math5 with l PGSIZE; autosolve.
rewrite H.
rewrite H0.
rewrite <- math5; autosolve.
Qed.

Lemma Addr_parts : forall l, l = (Pg l * PGSIZE + Off l)%Z.
intros.
unfold Pg, Off.
rewrite <- math5; autosolve.
Qed.

Lemma ValidPA_ValidVA : forall l, ValidPA l = true -> ValidVA l = true.
intros.
unfold ValidPA, ValidVA in *.
unfold NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.


Lemma ValidPA_ValidPg : forall l, ValidPA (l) = true -> ValidPg (Pg l) = true.
intros.
unfold ValidPA, ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold Pg in *.
assert (0 > l)%Z.
apply math6 with PGSIZE; autosolve.
autosolve.
Qed.

Lemma ValidVA_ValidVPg : forall l, ValidVA l = true -> ValidVPg (Pg l) = true.
intros.
unfold ValidVA, ValidVPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold Pg in *.
assert (0 > l)%Z.
apply math6 with PGSIZE; autosolve.
autosolve.
Qed.

Lemma ValidVA_Phys : forall addr, ValidVA addr = true -> isPhys addr = true -> ValidPA addr = true.
intros.
unfold isPhys in *.
unfold ValidVA, ValidPA, ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma ValidPg_ValidVPg : forall pg , ValidPg pg = true -> ValidVPg pg = true.
intros.
unfold ValidPg, ValidVPg in *.
unfold NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.


Lemma Off0 : forall x, (x + Off 0 = x)%Z.
intros.
unfold Off.
rewrite mod0; autosolve.
Qed.


Lemma ValidVPg_existsVA : forall pg, ValidVPg pg = true -> exists va, ValidVA va = true /\ Pg va = pg.
intros.
exists (pg * PGSIZE + Off 0)%Z.
rewrite Pg_simpl.
split; auto.
unfold ValidVPg, ValidVA in *.
rewrite Pg_simpl.
rewrite Off0.
unfold WordAligned.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
replace (pg * 4096)%Z with ((pg * 512) * 8)%Z in H3; autosolve.
rewrite mod_elim in H3; autosolve.
Qed.

Lemma WordAligned_off : forall pg l,
WordAligned l = true ->
WordAligned (pg * 4096 + Off l) = true.
intros.
unfold WordAligned in *.
rewrite mod_plus; autosolve.
replace (pg * 4096)%Z with ((pg * 512) * 8)%Z; autosolve.
rewrite mod_elim; autosolve.
simpl.
unfold Off.
rewrite Zmod_mod.
replace (4096)%Z with (512 * 8)%Z; autosolve.
rewrite (Z_div_exact_2 l 8); autosolve.
rewrite Zmult_comm.
rewrite Zmult_mod_distr_r.
rewrite Z_mod_mult; auto.
Qed.

Lemma ValidPA_parts : forall pg l, ValidPg pg = true -> ValidVA l = true -> ValidPA (pg * PGSIZE + Off l) = true.
intros.
unfold ValidPg, ValidVA, ValidPA in *.
rewrite Pg_simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply WordAligned_off; auto.
assert (0 <= pg * PGSIZE)%Z.
autosolve.
assert (0 <= Off l)%Z.
unfold Off.
apply mod_le; autosolve.
assert (0 <= pg * PGSIZE + Off l)%Z; autosolve.
Qed.

Lemma ValidPg_ValidPA_zero : forall pg, ValidPg (pg) = true -> ValidPA (pg * PGSIZE + Off 0) = true.
intros.
unfold ValidPg, ValidPA in *.
rewrite Pg_simpl.
unfold Off.
replace (0 mod 4096)%Z with 0%Z; autosolve.
replace (pg * PGSIZE + 0)%Z with (pg * PGSIZE)%Z; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold WordAligned.
replace (pg * 4096)%Z with (pg * 512 * 8)%Z; autosolve.
rewrite mod_elim; autosolve.
Qed.


Lemma ValidPA_separate : forall l l', ValidPA l = true -> ValidPA l' = false -> l <> l'.
intros.
red; intros.
rewrite H1 in *.
rewrite H in H0; autosolve.
Qed.

Lemma ValidVPg_ValidVA_zero : forall pg, ValidVPg pg = true -> ValidVA (pg * PGSIZE + Off 0)%Z = true.
intros.
unfold ValidVPg, ValidVA in *.
rewrite Pg_simpl.
unfold Off.
replace (0 mod 4096)%Z with 0%Z; autosolve.
replace (pg * PGSIZE + 0)%Z with (pg * PGSIZE)%Z; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold WordAligned.
replace (pg * 4096)%Z with (pg * 512 * 8)%Z; autosolve.
rewrite mod_elim; autosolve.
Qed.

Lemma ValidVAComp : forall pg l, ValidVPg pg = true -> ValidVA l = true -> ValidVA (pg * PGSIZE + Off l) = true.
intros.
unfold ValidVPg, ValidVA in *.
rewrite Pg_simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply WordAligned_off; auto.
assert (0 <= pg * PGSIZE)%Z.
autosolve.
assert (0 <= Off l)%Z.
unfold Off.
apply mod_le; autosolve.
assert (0 <= pg * PGSIZE + Off l)%Z; autosolve.
Qed.

Lemma InvalidPA_InvalidPg : forall l, ValidVA l = true -> ValidPA l = false -> ValidPg (Pg l) = false.
intros.
unfold ValidVA, ValidPA, ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; auto.
rewrite H in H0; auto.
Qed.

Definition PTROOT : Z := (352 * PGSIZE)%Z. (* 0x160 *)
Lemma PTROOT_Valid : 
  ValidPA PTROOT = true /\ ((PTROOT + VPAGES * 8) < MEMSIZE)%Z .
unfold PTROOT, ValidPA, MEMSIZE, Pg, NPAGES in *.
split; autosolve.
Qed.

Opaque NPAGES VPAGES PTROOT.
