(* hw.v *)
(* Definition of the HW memory model *)
(* TODO: remove TLB, as it is currently not completely implemented *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem.

Module HW_State <: Mem_T.

Record State := { data : Z -> Z; pe : bool; ptroot : Z; tlb : bool }.
Definition t := State.

Definition ValidVA (l :Z) : SCheck (State := State) := 
fun s => 
     if s.(pe) then
         ValidPA s.(ptroot) && 
         let ppg := s.(data) (s.(ptroot) + Pg(l) * 8) in
         ValidPA(ppg * PGSIZE + Off(l))
     else
	ValidPA l.

Definition getMemCheck (l:Z) : SCheck (State := State) := ValidVA l.
 
Definition getMem (l:Z) : SGet (State := State) Z :=
fun s => if s.(pe) then
                 let ppg := s.(data) (s.(ptroot) + Pg(l) * 8) in 
                 let pa := (ppg * PGSIZE + Off l)%Z in
                 s.(data) pa
              else
	  	 s.(data) l.

Definition setMemCheck (l : Z) : SCheck (State := State) := ValidVA l.
Definition setMem (l z : Z) :  SAct (State := State) :=
fun s => if s.(pe) then
              let ppg := s.(data) (s.(ptroot) + Pg(l) * 8) in 
              let pa := (ppg * PGSIZE + Off l)%Z in
              {| data := fun l' => if Zeq_bool pa l' then z else s.(data) l'; 
                      pe := s.(pe); ptroot := s.(ptroot); tlb := s.(tlb) |}
              else
              {| data := fun l' => if Zeq_bool l l' then z else s.(data) l'; 
                      pe := s.(pe); ptroot := s.(ptroot); tlb := s.(tlb) |}.
              

Definition setPECheck (root :Z ) : SCheck (State := State) :=
fun s => true.
Definition setPE  : SAct (State := State) := 
fun s => {| data := s.(data); pe := true; ptroot := s.(ptroot); tlb := true |}.


Definition setPTROOTCheck (root :Z ) : SCheck (State := State) :=
fun s => ValidPA root.
Definition setPTROOT (root : Z) : SAct (State := State) := 
fun s => {| data := s.(data); pe := s.(pe); ptroot := root; tlb := true |}.

Definition getTLB : SGet (State := State) bool :=
fun s => s.(tlb).

Definition setTLB (t : bool) :  SAct (State := State) :=
fun s =>  {| data := s.(data); pe := s.(pe); ptroot := s.(ptroot); tlb := t |}.


(* getMem simplifications *)
Lemma setMem_unneeded : forall l z s, getMemCheck l s = true -> (getMem l s = z) -> (setMem l z s = s).
intros.
unfold getMem, setMem, ValidVA, ValidPA in *; simpl; subst.
makenew' (pe s); simpl in *.

destruct s; simpl in *.
f_equal; simpl; autosolve.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.

destruct s; simpl in *.
f_equal; simpl; autosolve.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.
Qed.

End HW_State. 
