(* mem_sig.v *)
(* The specifications for the procedures of memory allocator (over the PE and PD models) *)
(* Also includes the stub for kernel_init in the PE model *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.

Require Import spec.
Require Import label.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.
Require Import Program.

Require Import c.
Require Import pe.
Require Import ale.
(* Require Import pm. *)
Require Import pe_lib.

Require Import sact.
Require Import actsmp.

Opaque PE_State.PMM_data_loc.

Module mem_sig.
Import C_PE SCAP.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).
	
Definition allocpg (pg : Z) : State -> State -> Prop :=
  fun s s' => s'.(memory) = setPageInfo pg true s.(memory) /\ StackRets pg s.(stack) s'.(stack).

Definition mem_alloc_spec : Action State :=
{|
pre := (fun s => 
      StackCheck (getArgsCheck 0) s = true /\ StackCheck isFrame s = true /\ 
      MemGet (getTLB) s = true /\ DM s.(memory) 
     );

rel := (fun s s' =>
   (exists page, (1 <= page < NPAGES)%Z /\  MemGet (getPageInfo page) s = false /\ allocpg page s s')
   \/
   (s' = StackAct (pop +++ setRet 0%Z) s))
|}.


Definition mem_free_spec : Action State :=
{| pre := (fun s => StackCheck (getArgsCheck 1) s = true /\ 
                  let pg := nth 0 (StackGet getArgs s) 0%Z in 
                  (0 < pg < NPAGES)%Z /\
                  MemGet (getPageInfo pg) s = true /\
                  MemGet (getTLB) s = true /\ DM s.(memory) 
	      	);

     rel := (fun s s' => let pg := nth 0 (StackGet getArgs s) 0%Z in
                 if MemGet (getPageInfo pg) s then
                         s' = (StackAct (pop +++ setRet 0%Z)  +++ (MemAct (setPageInfo pg false))) s
                 else False
               )
|}.



Definition mem_spec : SpecHeap := fun l =>
  match l with
  | mem_free_label => Some mem_free_spec
  | mem_alloc_label => Some mem_alloc_spec
  | _ => None
  end.




Definition PE_kernel_init_spec : Action C_PE.State.t := {|
pre := (fun s => 
     C_PE.State.StackCheck (C_PM.State.Stack.getArgsCheck 0) s = true /\
     (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                      s.(C_PE.State.memory).(PE_State.data) (PMM_data_loc + Pg l * 8) = 1%Z) /\
     (forall l, ValidPA l = true -> PMMDom l = true -> 
                      s.(C_PE.State.memory).(PE_State.data) (PMM_data_loc + Pg l * 8) = 1%Z) /\
     (forall pg, ValidPg pg = true -> s.(C_PE.State.memory).(PE_State.data) (PTROOT + pg * 8) = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s.(C_PE.State.memory).(PE_State.data) (PTROOT + pg * 8)= 0%Z) /\
      s.(C_PE.State.memory).(PE_State.tlb) = true
      );
rel := (fun s s' => False)
|}.

Lemma lib_mem_disj : C_PE.SCAP.SpecHeap.SpecHeap_disjoint PE_Lib.pe_lib_spec mem_sig.mem_spec.
red; intros.
destruct l; autosolve.
Qed.



End mem_sig.

