(* pm.v *)
(* Definition of the PD memory model *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem.

Module PM_State <: Mem_T.

Definition State := Z -> Z.

Definition t := State.

Definition getMemCheck (l:Z) : SCheck (State := State) := fun s => ValidPA l.
Definition getMem (l:Z) : SGet (State := State) Z := fun s => s l.

Definition setMemCheck (l : Z) : SCheck (State := State) := fun s => ValidPA l.
Definition setMem (l z : Z) :  SAct (State := State) :=
  fun s => (fun l' => if Zeq_bool l l' then z else s l').

(* getMem simplifications *)

Lemma setMem_unneeded : forall l z s, getMemCheck l s = true -> (getMem l s = z) -> (setMem l z s = s).
intros.
unfold getMem, setMem, ValidPA in *; simpl; subst.
repeat autoclean.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.
Qed.

Lemma getMem_setMem : 
  forall l l' z s, getMem l (setMem l' z s) = if Zeq_bool l l' then z else getMem l s.
intros.
unfold getMem, setMem.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
Qed.

Hint Rewrite getMem_setMem : mem.

End PM_State. 
