(* codeheap.v *)
(* Defines the notion of code heaps used to store procs *)


Require Import Equalities.
Require Import Logic.FunctionalExtensionality.
Require Import label.
Require Import zbool_assist.

Module Type CodeHeap_t (p : Typ).

Definition CodeHeap := Label -> option (p.t).

Definition CodeHeap_empty : CodeHeap := fun l => None.

Definition CodeHeap_join (C C' : CodeHeap) : CodeHeap :=
  fun l => match C l with
	   | Some g => Some g
	   | None => C' l
	   end.

Parameter CodeHeap_join_empty : forall C, CodeHeap_join CodeHeap_empty C = C.

Definition CodeHeap_disjoint (C C' : CodeHeap) : Prop :=
 forall l, 
  match C l with
  | Some _ => C' l = None
  | _ => True
  end.
End CodeHeap_t.


Module CodeHeap (p : Typ) <: CodeHeap_t p.

Definition CodeHeap := Label -> option (p.t).

Definition CodeHeap_empty : CodeHeap := fun l => None.

Definition CodeHeap_join (C C' : CodeHeap) : CodeHeap :=
  fun l => match C l with
	   | Some g => Some g
	   | None => C' l
	   end.

Require Import Logic.FunctionalExtensionality.

Lemma CodeHeap_join_empty : forall C, CodeHeap_join CodeHeap_empty C = C.
intros.
apply functional_extensionality; firstorder.
Qed.

Definition CodeHeap_disjoint (C C' : CodeHeap) : Prop :=
 forall l, 
  match C l with
  | Some _ => C' l = None
  | _ => True
  end.

End CodeHeap.