(* label.v
 * This is an inductive type of labels used by the BabyVMM
 * Ideally, labels should be defined as an abstract type, and the specific constructors moved to the particular code,
 * however, this would require some kind of additional modularity
 * Or alternatively, we could use a decidable type such as nat - but that complicates case analysis
 * Thus, this approach is a compromise.
 *)

Inductive Label :=
kernel_label |
kernel_init_label |
hw_init_label |
(* tlb_reset_label | *) set_ptroot_label | pe_enable_label |
pmap_set_label | 
pmap_lookup_label | 
as_reserve_label |
as_release_label |

mem_alloc_label | mem_incref_label | mem_decref_label | mem_free_label | mem_pi2ptr_label 
| mem_init_label | mem_init_loop1_label | mem_init_loop2_label | mem_init_loop3_label
| mem_init_reserveself_label
| mem_alloc_loop_label | mem_alloc_body_label
| bzero_label 
| pmap_new_helper_label | table_free_label | table_loop_free_label | table_entry_free_label
| pmap_new_label | pmap_init_label | pmap_init_loop1_label | pmap_init_loop2_label
| pmap_level_ind_label | pmap_walk_label | pmap_walk_helper_label 
| pmap_walk_helper1_label | pmap_walk_helper2_label | pmap_addtable_label
| stack_push_fail_label
| pmap_update_label.

Definition tlb_reset_label := pe_enable_label.
