(* hw_init.v *)
(* The definition and certification of the code of the hw_init procedure *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import spec.


Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import hw.
Require Import hw_lib.

Module hw_init.
Import C_HW State Mem Stack Ops SCAP Process.

Require Import sact.
Require Import actsmp.

(***************************************************)

Require Import label.
Require Import pe.
Require Import ale.

Opaque PTROOT PE_State.PTROOT ALE_State.PTROOT PE_State.PMM_data_loc.

Definition Psi (f : Label) : option (Action State) := 
   match f with
   | kernel_init_label => Some HW_Lib.HW_kernel_init_spec
   | pe_enable_label => Some HW_Lib.pe_enable_spec
   | set_ptroot_label => Some HW_Lib.set_ptroot_spec
   | mem_init_label => Some HW_Lib.HW_mem_init_spec
   | pmap_init_label => Some HW_Lib.HW_pmap_init_spec
   | _ => None
   end.

Definition junk_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Axiom mem_init_flabel : FLABEL.
Axiom pe_enable_flabel : FLABEL.
Axiom set_ptroot_flabel : FLABEL.
Axiom kernel_init_flabel : FLABEL.
Axiom pmap_init_flabel : FLABEL.

Definition hw_init : Process :=
  Iperf (perf_fcall mem_init_flabel nil) :::
  Icall mem_init_label :::
  Iperf (perf_readret junk_var) :::
  Iperf (perf_fcall pmap_init_flabel nil) :::
  Icall pmap_init_label :::
  Iperf (perf_readret junk_var) :::
  Iperf (perf_fcall set_ptroot_flabel (expr_z PTROOT :: nil)) :::
  Icall set_ptroot_label :::
  Iperf (perf_readret junk_var) :::
  Iperf (perf_fcall pe_enable_flabel nil) :::
  Icall pe_enable_label :::
  Iperf (perf_readret junk_var) :::
  Iperf (perf_fcall kernel_init_flabel nil) :::
  Icall kernel_init_label :::
  Iperf (perf_readret junk_var)  
  .

Definition hw_init_spec : Action State :=
 {| pre := (fun (s : State) => StackCheck (getArgsCheck 0) s = true /\ 
                        s.(C_HW.State.memory).(HW_State.pe) = false); 
   rel := fun (s s' : State) => False |}. 

  
Definition hw_init_spec' : Action State :=
  Actions (perf_fcall mem_init_flabel nil) ;;
  HW_Lib.HW_mem_init_spec ;;
  Actions (perf_readret junk_var) ;;
  Actions (perf_fcall mem_init_flabel nil) ;;
  HW_Lib.HW_pmap_init_spec ;;
  Actions (perf_readret junk_var) ;;
  Actions (perf_fcall set_ptroot_flabel (expr_z PTROOT :: nil)) ;;
  HW_Lib.set_ptroot_spec ;;
  Actions (perf_readret junk_var) ;;
  Actions (perf_fcall pe_enable_flabel nil) ;;
  HW_Lib.pe_enable_spec ;;
  Actions (perf_readret junk_var) ;;
  Actions (perf_fcall kernel_init_flabel nil) ;;
  HW_Lib.HW_kernel_init_spec ;;
  Actions (perf_readret junk_var)  
  .

Definition WF_hw_init' : WFProc expr_to_Prop Mc_Machine Psi hw_init_spec' hw_init.
unfold hw_init_spec'; unfold hw_init; auto 999.
Qed.

Require ale_pe.

Definition WF_hw_init : WFProc expr_to_Prop Mc_Machine Psi hw_init_spec hw_init.
apply WFProc_weaken with hw_init_spec'; [|apply WF_hw_init'].
unfold hw_init_spec, hw_init_spec'.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.

split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

unfold StackCheck, getRetCheck in *; simpl in *.
repeat split; repeat autoclean; simpl in *; try repeat rewriter; simpl in *; autosolve.

apply H6; auto.
rewrite H4; simpl; auto.
apply H5; auto.

rewrite H12; autosolve.
apply H17; auto.
apply ValidPA_ValidPg; auto.

rewrite H12; autosolve.
apply H18; auto.
apply ValidPA_ValidPg; auto.

apply H10; auto.
apply H11; auto.

(* rel *)
repeat autoclean; autosolve.
Qed.

Definition hw_init_lib : C_HW.SCAP.CodeHeap := fun f =>
   match f with
       | hw_init_label => Some hw_init
	| _ => None
   end.

Definition hw_init_dep := Psi.

Definition hw_init_lib_spec : C_HW.SCAP.SpecHeap.SpecHeap := fun f =>
   match f with
       | hw_init_label => Some hw_init_spec
	| _ => None
   end.

Lemma WF_hw_init_lib : C_HW.SCAP.WFCode expr_to_Prop C_HW.Mc_Machine hw_init_dep hw_init_lib hw_init_lib_spec.
assert (SpecHeap_disjoint hw_init_dep hw_init_lib_spec).
red; intros; destruct l; autosolve.
apply C_HW.SCAP.WFCode_1 with H.
intros.
destruct f; simpl in *; autosolve.
exists hw_init; split; auto.
red in H0; simpl in *.
inversion H0; subst; clear H0.
apply WFProc_strengthen with hw_init_dep.
red; intros; destruct l; simpl; autosolve; try apply ActionGTE_refl.
apply WF_hw_init.
Qed.

End hw_init.