(* main.v *)
(* Links all the certified modules together to get a complete certified module containing the entire VMM *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.

Require Import spec.
Require Import label.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.
Require Import Program.

Require Import c.
Require Import "as".
Require Import as_lib.

Require Import sact.
Require Import actsmp.



(*******************************************************************************)

Require Import pmap.
Require Import pmap_lib.
Require Import as_sig.
Require Import as_lib_cert.
Require Import as_pmap.

(* Kernel can be refined to PMAP and linked with C^{as} *)

Lemma WF_as_pmap : forall kernel_code kernel_spec
         (disj : C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint (C_AS_PMAP.PsiMake kernel_spec) as_sig.as_spec)
         (disj' : C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint (C_AS_PMAP.PsiMake kernel_spec) PMAP_Lib.pmap_lib_spec),
         C_PMAP.SCAP.CH.CodeHeap_disjoint as_lib_code (C_AS_PMAP.CMake kernel_code) ->
         C_AS.SCAP.WFCode C_AS.State.expr_to_Prop C_AS.Mc_Machine AS_Lib.as_lib_spec kernel_code kernel_spec ->

      C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine 
                PMAP_Lib.pmap_lib_spec 
                (C_PMAP.SCAP.CH.CodeHeap_join (C_AS_PMAP.CMake kernel_code) as_lib_code) 
                (C_PMAP.SCAP.SpecHeap.SpecHeap_join (C_AS_PMAP.PsiMake kernel_spec) as_sig.as_spec disj).
intros.
apply as_pmap.WF_kernel_pmap'; auto.
apply WF_as_lib.
Qed.

(****************************************************************)

(* Refining PMAP to ALE *)

Require Import ale.
Require Import ale_lib.
Require Import pmap_sig.
Require Import pmap_ale.
Require Import pmap_lib_cert.

Lemma WF_pmap_ale : forall C PSI
         (disj : C_ALE.SCAP.SpecHeap.SpecHeap_disjoint (C_PMAP_ALE.PsiMake PSI) pmap_sig.pmap_spec)
         (disj' : C_ALE.SCAP.SpecHeap.SpecHeap_disjoint (C_PMAP_ALE.PsiMake PSI) ALE_Lib.ale_lib_spec),
         C_ALE.SCAP.CH.CodeHeap_disjoint pmap_lib_code (C_PMAP_ALE.CMake C) ->
         C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine PMAP_Lib.pmap_lib_spec C PSI ->

      C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine 
                ALE_Lib.ale_lib_spec 
                (C_ALE.SCAP.CH.CodeHeap_join (C_PMAP_ALE.CMake C) pmap_lib_code) 
                (C_ALE.SCAP.SpecHeap.SpecHeap_join (C_PMAP_ALE.PsiMake PSI) pmap_sig.pmap_spec disj).
intros.
apply pmap_ale.WF_kernel_pmap'; auto.
apply WF_pmap_lib.
Qed.

(****************************************************************)

(* ALE -> PE *)

Require Import pe.
Require Import pe_lib.
Require Import mem_sig.
Require Import mem_sig_ald.
Require Import mem_lib_cert.
Require Import ale_pe.

Lemma WF_ale_pe : forall C PSI
         (disj : C_PE.SCAP.SpecHeap.SpecHeap_disjoint (C_ALE_PE.PsiMake PSI) mem_lib_cert.mem_spec)
         (disj' : C_PE.SCAP.SpecHeap.SpecHeap_disjoint (C_ALE_PE.PsiMake PSI) PE_Lib.pe_lib_spec),
         C_PE.SCAP.CH.CodeHeap_disjoint mem_code (C_ALE_PE.CMake C) ->
         C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine ALE_Lib.ale_lib_spec C PSI ->

      C_PE.SCAP.WFCode C_PE.State.expr_to_Prop C_PE.Mc_Machine 
                PE_Lib.pe_lib_spec 
                (C_PE.SCAP.CH.CodeHeap_join (C_ALE_PE.CMake C) mem_code) 
                (C_PE.SCAP.SpecHeap.SpecHeap_join (C_ALE_PE.PsiMake PSI) mem_lib_cert.mem_spec disj).
intros.
apply ale_pe.WF_kernel_ale'; auto.
apply lib_mem_disj.
red; intros.
destruct l; autosolve.
apply WF_mem_lib.
Qed.

(****************************************************************)

(* PE -> HW *)

Require Import hw.
Require Import hw_lib.
Require Import pe_hw.

Lemma WF_pe_hw : forall C PSI
         (disj' : C_HW.SCAP.SpecHeap.SpecHeap_disjoint (C_PE_HW.PsiMake PSI) HW_Lib.hw_lib_spec),
         C_PE.SCAP.WFCode C_PE.State.expr_to_Prop C_PE.Mc_Machine PE_Lib.pe_lib_spec C PSI ->

      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine 
                HW_Lib.hw_lib_spec 
                (C_PE_HW.CMake C) 
                (C_PE_HW.PsiMake PSI).
intros.
apply pe_hw.WF_kernel_pe'; auto.
Qed.

(******************************************************************)

(* ALD -> PD *)

Require Import ald.
Require Import ald_pm.

Lemma WF_ald_pm : forall C PSI,
      C_ALD.SCAP.WFCode C_ALD.State.expr_to_Prop C_ALD.Mc_Machine C_ALD.SCAP.SpecHeap.SpecHeap_empty C PSI ->
      C_PM.SCAP.WFCode C_PM.State.expr_to_Prop C_PM.Mc_Machine 
                C_PM.SCAP.SpecHeap.SpecHeap_empty 
                (C_ALD_PM.CMake C)
                (C_ALD_PM.PsiMake PSI).
intros.
apply ald_pm.WF_ALD_pm; auto.
Qed.

(*******************************************************************)

(* PD -> HW *)

Require Import pm_hw.

Lemma WF_pm_hw : forall C PSI,
      C_PM.SCAP.WFCode C_PM.State.expr_to_Prop C_PM.Mc_Machine C_PM.SCAP.SpecHeap.SpecHeap_empty C PSI ->
      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine 
                C_HW.SCAP.SpecHeap.SpecHeap_empty 
                (C_PM_HW.CMake C)
                (C_PM_HW.PsiMake PSI).
intros.
apply pm_hw.WF_PM_HW; auto.
Qed.

(*******************************************************************)
(*******************************************************************)
(*******************************************************************)
(*******************************************************************)
(*******************************************************************)
(*******************************************************************)
(*******************************************************************)
(*******************************************************************)
(*******************************************************************)

Require Import mem_init.
Require Import pmap_init.
Require Import hw_init.

(******************************************************)
(* Refine the mem_init and pt_init to the HW machine *)
(* C1 - resulting code, PSI1 - resulting spec *)
(* WF1 - proof of well-formedness of resulting specs and code *)

Definition C1 : C_HW.SCAP.CH.CodeHeap := 
    C_HW.SCAP.CH.CodeHeap_join 
              (C_PM_HW.CMake mem_init.mem_init_CODE) 
              (C_PM_HW.CMake (C_ALD_PM.CMake pmap_init.pmap_init_CODE)).

Lemma disj1 : C_HW.SCAP.SpecHeap.SpecHeap_disjoint
              (C_PM_HW.PsiMake mem_init.mem_init_PSI) 
              (C_PM_HW.PsiMake (C_ALD_PM.PsiMake pmap_init.pmap_init_PSI)).
red; intros.
destruct l; simpl; auto.
Qed.

Definition PSI1 : C_HW.SCAP.SpecHeap.SpecHeap := 
    C_HW.SCAP.SpecHeap.SpecHeap_join 
              (C_PM_HW.PsiMake mem_init.mem_init_PSI) 
              (C_PM_HW.PsiMake (C_ALD_PM.PsiMake pmap_init.pmap_init_PSI))
              disj1.

Lemma WF1 : 
      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine HW_Lib.hw_lib_spec C1 PSI1.
unfold C1, PSI1.
apply C_HW.SCAP.WFCode_stronghypo with (C_HW.SCAP.SpecHeap.SpecHeap_empty).
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.
apply C_HW.SCAP.WFCode_join.
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.
apply WF_pm_hw.
apply mem_init.WF_mem_init_lib.
apply WF_pm_hw.
unfold C_ALD_PM.CMake.
apply WF_ald_pm.
apply pmap_init.WF_pmap_init_lib.
Qed.


(*********************************************************)
(* We assume a certified kernel *)

Definition kernel_proc_spec : Action C_AS.State.t := {|
pre := (fun s => C_AS.State.StackCheck (C_AS.State.Stack.getArgsCheck 0) s = true);
rel := (fun s s' => False)
|}.

Axiom kernel : C_AS.SCAP.Process.Process.
Definition kernel_code : C_AS.SCAP.CodeHeap := fun l =>
  match l with
  | kernel_init_label => Some kernel
  | _ => None
  end.

Definition kernel_spec : C_AS.SCAP.SpecHeap.SpecHeap := fun l =>
  match l with
  | kernel_init_label => Some kernel_proc_spec
  | _ => None
  end.
Axiom WF_kernel : C_AS.SCAP.WFCode C_AS.State.expr_to_Prop C_AS.Mc_Machine 
                              AS_Lib.as_lib_spec kernel_code kernel_spec.

(******************************************************************)
(* A few annoying proofs that the domains of various code heaps are disjoint *)

Lemma disj2 : C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint
                                (C_AS_PMAP.PsiMake kernel_spec)
                                as_sig.as_spec.
red; intros; destruct l; simpl; autosolve.
Qed.

Lemma disj3 : 
(C_ALE.SCAP.SpecHeap.SpecHeap_disjoint
                 (C_PMAP_ALE.PsiMake 
                      (C_PMAP.SCAP.SpecHeap.SpecHeap_join
                                (C_AS_PMAP.PsiMake kernel_spec)
                                as_sig.as_spec
                                disj2
                      )
                 )
                 pmap_sig.pmap_spec).
red; intros; destruct l; simpl; auto.
Qed.

Lemma disj4 : 
(C_PE.SCAP.SpecHeap.SpecHeap_disjoint
          (C_ALE_PE.PsiMake 
                 (C_ALE.SCAP.SpecHeap.SpecHeap_join
                 (C_PMAP_ALE.PsiMake 
                      (C_PMAP.SCAP.SpecHeap.SpecHeap_join
                                (C_AS_PMAP.PsiMake kernel_spec)
                                as_sig.as_spec
                                disj2
                      )
                 )
                 pmap_sig.pmap_spec
                 disj3
                 )
          )
          (* mem_sig. *)mem_spec
).
red; intros; destruct l; simpl; auto.
Qed.

(*********************************************************)
(* Refine the kernel from AS to HW, linking it with all the relevant modules *)
(* PSI2 - is the refined and linked spec, C2 is the code of all the modules put together *)
(* WF2 - proof of well-formedness of the refined modules *)

Definition PSI2 : C_HW.SCAP.SpecHeap.SpecHeap :=
    (C_PE_HW.PsiMake 
          (C_PE.SCAP.SpecHeap.SpecHeap_join
          (C_ALE_PE.PsiMake 
                 (C_ALE.SCAP.SpecHeap.SpecHeap_join
                 (C_PMAP_ALE.PsiMake 
                      (C_PMAP.SCAP.SpecHeap.SpecHeap_join
                                (C_AS_PMAP.PsiMake kernel_spec)
                                as_sig.as_spec
                                disj2
                      )
                 )
                 pmap_sig.pmap_spec
                 disj3
                 )
          )
          mem_spec
          disj4
          )
     ).

Definition C2 : C_HW.SCAP.CH.CodeHeap :=
    (C_PE_HW.CMake 
         (C_PE.SCAP.CH.CodeHeap_join
         (C_ALE_PE.CMake
              (C_ALE.SCAP.CH.CodeHeap_join
              (C_PMAP_ALE.CMake   
                  (C_PMAP.SCAP.CH.CodeHeap_join (C_AS_PMAP.CMake kernel_code) as_lib_code))
              pmap_lib_code
              )
         )
         mem_code
         )
   ).

Lemma WF2 : 
      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine HW_Lib.hw_lib_spec C2 PSI2.
apply WF_pe_hw.
red; intros; destruct l; simpl; auto.
apply WF_ale_pe.
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.
apply WF_pmap_ale.
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.
apply WF_as_pmap.
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.
exact WF_kernel.
Qed.

(************************************************************)
(* Link C1 and C2 - and guarantee well-formedness *)


Definition C3 := C_HW.SCAP.CH.CodeHeap_join C1 C2.
Lemma PSI1_PSI2_disj : C_HW.SCAP.SpecHeap.SpecHeap_disjoint PSI1 PSI2.
red; intros; destruct l; simpl; auto.
Qed.
Definition PSI3 := C_HW.SCAP.SpecHeap.SpecHeap_join PSI1 PSI2 PSI1_PSI2_disj.

Lemma WF3 : 
      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine HW_Lib.hw_lib_spec C3 PSI3.
apply C_HW.SCAP.WFCode_join.
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.
apply WF1.
apply WF2.
Qed.

(**************************************)
(* Strengthen the init code to use the refined specifications *)

Definition HW_init_dep' := fun f => 
   match f with
   | kernel_init_label => Some
(C_PE_HW.RF.reprAction C_PE_HW.CRepr
  (C_ALE_PE.RF.reprAction C_ALE_PE.CRepr
     (C_PMAP_ALE.RF.reprAction C_PMAP_ALE.CRepr
        (C_AS_PMAP.RF.reprAction C_AS_PMAP.CRepr kernel_proc_spec))))
   | pe_enable_label => Some HW_Lib.pe_enable_spec
   | set_ptroot_label => Some HW_Lib.set_ptroot_spec
   | mem_init_label => Some
(C_PM_HW.RF.reprAction C_PM_HW.CRepr mem_sig_ald.PM_mem_init_spec)
   | pmap_init_label => Some
(C_PM_HW.RF.reprAction C_PM_HW.CRepr
  (C_ALD_PM.RF.reprAction C_ALD_PM.CRepr ald_lib.ALD_Lib.ALD_pmap_init_spec))
   | _ => None
   end.

Lemma WF_hw_init_lib' : C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine 
                                    HW_init_dep' hw_init.hw_init_lib hw_init.hw_init_lib_spec.
apply C_HW.SCAP.WFCode_stronghypo with hw_init.hw_init_dep.
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.

apply ActionGTE_trans with (C_PE_HW.RF.reprAction C_PE_HW.CRepr mem_sig.PE_kernel_init_spec).
apply pe_hw.kernel_init_weak.
apply C_PE_HW.RF.ActionGTE_repr.
apply ActionGTE_trans with (C_ALE_PE.RF.reprAction C_ALE_PE.CRepr pmap_sig.ALE_kernel_init_spec).
apply ale_pe.kernel_init_weak.
apply C_ALE_PE.RF.ActionGTE_repr.
apply ActionGTE_trans with (C_PMAP_ALE.RF.reprAction C_PMAP_ALE.CRepr as_sig.PMAP_kernel_init_spec).
apply pmap_ale.kernel_init_weak.
apply C_PMAP_ALE.RF.ActionGTE_repr.
apply as_pmap.kernel_init_weak.

apply pm_hw.mem_init_weak.

apply ActionGTE_trans with (C_PM_HW.RF.reprAction C_PM_HW.CRepr mem_sig_ald.PM_pmap_init_spec).
apply pm_hw.pmap_init_weak.
apply C_PM_HW.RF.ActionGTE_repr.
apply ald_pm.pmap_init_weak.

apply hw_init.WF_hw_init_lib.
Qed.

(******************************************)
(* Combine the C3 code heap with intialization *)
(* C4 now contains the VMM including init, and our kernel hypothesis *)
(* PSI4 is the specifications of these components in the HW model *)
(* WF4 is the proof that C4 is well formed under PSI4, completing the proof *)

Definition C4 := C_HW.SCAP.CH.CodeHeap_join C3 hw_init.hw_init_lib.
Lemma PSI3_hw_init_disj : C_HW.SCAP.SpecHeap.SpecHeap_disjoint PSI3 hw_init.hw_init_lib_spec.
red; intros; destruct l; simpl; auto.
Qed.
Definition PSI4 := C_HW.SCAP.SpecHeap.SpecHeap_join PSI3 hw_init.hw_init_lib_spec PSI3_hw_init_disj.

Lemma compat1 : (C_HW.SCAP.SpecHeap.SpecHeap_compat HW_Lib.hw_lib_spec hw_init.hw_init_lib_spec).
red; intros; destruct l; simpl; auto.
Qed.

Lemma compat2 : (C_HW.SCAP.SpecHeap.SpecHeap_compat HW_init_dep' PSI3).
red; intros; destruct l; simpl; auto.
Qed.


Lemma cdisj : C_HW.SCAP.CH.CodeHeap_disjoint C3 hw_init.hw_init_lib.
red; intros; destruct l; simpl; auto.
Qed.


Lemma compat3 : C_HW.SCAP.SpecHeap.SpecHeap_compat HW_Lib.hw_lib_spec HW_init_dep'.
red; intros; destruct l; simpl; auto.
Qed.


Lemma WF4 : 
      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine HW_Lib.hw_lib_spec C4 PSI4.
unfold C4.
unfold PSI4.
apply C_HW.SCAP.WFCode_stronghypo with
(C_HW.SCAP.SpecHeap.SpecHeap_add 
            (C_HW.SCAP.SpecHeap.SpecHeap_sub HW_Lib.hw_lib_spec hw_init.hw_init_lib_spec) 
            (C_HW.SCAP.SpecHeap.SpecHeap_sub HW_init_dep' PSI3)).
red; intros; destruct l; simpl; auto.
red; intros; destruct l; simpl; auto.
apply (C_HW.SCAP.WFCode_linking C_HW.State.expr_to_Prop compat3 PSI3_hw_init_disj cdisj compat1 compat2).
apply WF3.
apply WF_hw_init_lib'.
Qed.

(***********************************************************)