(* mem_init.v *)
(* Definition and certification of mem_init procedure *)
(* Includes code and certification 3 subprocedures for loops *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.

Require Import spec.
Require Import scap.


Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.
Require Import Program.
Opaque impl.

Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import label.

Module mem_init.

Import C_PM State Mem Stack Ops SCAP Process.

(* Sact actions and structured simplification *)
Require Import sact.

(* Structured tacticts for regular actions - move to spec? *)
Require Import actsmp.
(***************************************************)

Definition i_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Require Import pe.
Require Import ale.
Require Import mem_sig_ald.

Opaque PTROOT ALE_State.PTROOT PE_State.PTROOT PE_State.PMM_data_loc.

Definition mem_init_loop1_spec : Action State := 
 {| pre := (fun (s : State) => StackCheck (getVarCheck i_var) s = true /\ 
                                           ValidPg (StackGet (getVar i_var) s) = true /\
                                           (StackGet (getVar i_var) s <= 100)%Z); 
   rel := fun (s s' : State) => let i := StackGet (getVar i_var) s  in
                            (forall pg, (i <= pg < 100)%Z -> (s'.(memory) (PE_State.PMM_data_loc + pg* 8)%Z = 0%Z)) /\
                            (forall l, ((l < PE_State.PMM_data_loc + i  * 8)%Z \/ (l >= PE_State.PMM_data_loc + 100*8)%Z) -> 
                                            s.(memory) l = s'.(memory) l) /\
                            (s'.(stack) = setVar i_var 100%Z s.(stack)) |}. 

Definition mem_init_loop2_spec : Action State := 
 {| pre := (fun (s : State) => StackCheck (getVarCheck i_var) s = true /\ 
                                           ValidPg (StackGet (getVar i_var) s) = true /\
                                           (StackGet (getVar i_var) s <= 512)%Z); 
   rel := fun (s s' : State) => let i := StackGet (getVar i_var) s  in
                            (forall pg, (i <= pg < 512)%Z -> (s'.(memory) (PE_State.PMM_data_loc + pg* 8)%Z = 1%Z)) /\
                            (forall l, ((l < PE_State.PMM_data_loc + i  * 8)%Z \/ (l >= PE_State.PMM_data_loc + 512*8)%Z) -> 
                                            s.(memory) l = s'.(memory) l) /\
                            (s'.(stack) = setVar i_var 512%Z s.(stack)) |}. 


Definition mem_init_loop3_spec : Action State := 
 {| pre := (fun (s : State) => StackCheck (getVarCheck i_var) s = true /\ 
                                           (ValidPg (StackGet (getVar i_var) s) = true \/ (StackGet (getVar i_var) s = NPAGES)) /\
                                           (StackGet (getVar i_var) s <= NPAGES)%Z); 
   rel := fun (s s' : State) => let i := StackGet (getVar i_var) s  in
                            (forall pg, (i <= pg < NPAGES)%Z -> (s'.(memory) (PE_State.PMM_data_loc + pg* 8)%Z = 0%Z)) /\
                            (forall l, ((l < PE_State.PMM_data_loc + i  * 8)%Z \/ (l >= PE_State.PMM_data_loc + NPAGES*8)%Z) -> 
                                            s.(memory) l = s'.(memory) l) /\
                            (s'.(stack) = setVar i_var NPAGES%Z s.(stack)) |}. 


Definition mem_init : Process :=
   Iperf (perf_readargs []) :::
   Iperf (perf_assign i_var (expr_z 1%Z)) :::
   Iperf (perf_store (expr_z PE_State.PMM_data_loc) (expr_z 0%Z)) :::
   Icall mem_init_loop1_label :::
   Icall mem_init_loop2_label :::
   Icall mem_init_loop3_label :::
   Iperf (perf_ret (expr_z 0%Z))
.

Definition if1 := expr_nomem_binop LT (expr_nomem_var i_var) (expr_nomem_z 100%Z).
Definition if2 := expr_nomem_binop LT (expr_nomem_var i_var) (expr_nomem_z 512%Z).
Definition if3 := expr_nomem_binop LT (expr_nomem_var i_var) (expr_nomem_z NPAGES).

Definition PMM_expr (pg : expr) : expr :=
  (expr_binop PLUS (expr_z PE_State.PMM_data_loc) (expr_binop TIMES pg (expr_z 8%Z))).


Definition mem_init_loop1 : Process :=
  Iplus (if1)
  (Iperf (perf_store (PMM_expr (expr_var i_var)) (expr_z 0%Z)) :::
   Iperf (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) :::
   Icall mem_init_loop1_label) 
  Inil.

Definition mem_init_loop2 : Process :=
  Iplus (if2)
  (Iperf (perf_store (PMM_expr (expr_var i_var)) (expr_z 1%Z)) :::
   Iperf (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) :::
   Icall mem_init_loop2_label) 
  Inil.

Definition mem_init_loop3 : Process :=
  Iplus (if3)
  (Iperf (perf_store (PMM_expr (expr_var i_var)) (expr_z 0%Z)) :::
   Iperf (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) :::
   Icall mem_init_loop3_label)   
  Inil.


Definition Psi (f : Label) : option (Action State) := 
   match f with
   | mem_init_loop1_label  => Some (mem_init_loop1_spec)
   | mem_init_loop2_label => Some (mem_init_loop2_spec)
   | mem_init_loop3_label => Some (mem_init_loop3_spec)
   | _ => None
   end.

Definition mem_init_spec' : Action State :=
   Actions (perf_readargs []) ;;
   Actions (perf_assign i_var (expr_z 1%Z)) ;;
   Actions (perf_store (expr_z PE_State.PMM_data_loc) (expr_z 0%Z)) ;;
   mem_init_loop1_spec ;;
   mem_init_loop2_spec ;;
   mem_init_loop3_spec ;;
   Actions (perf_ret (expr_z 0%Z))
.
Definition WF_mem_init' : WFProc expr_to_Prop Mc_Machine Psi mem_init_spec' mem_init.
unfold mem_init_spec'; unfold mem_init. auto 999.
Qed.


Definition mem_init_loop1_spec' : Action State :=
  ActionPlus (expr_to_Prop if1)
  (Actions (perf_store (PMM_expr (expr_var i_var)) (expr_z 0%Z)) ;;
   Actions (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) ;;
   mem_init_loop1_spec) 
  (ActionID State).


Definition WF_mem_init_loop1' : WFProc expr_to_Prop Mc_Machine Psi mem_init_loop1_spec' mem_init_loop1.
unfold mem_init_loop1_spec'; unfold mem_init_loop1. auto 999.
Qed.


Definition mem_init_loop3_spec' : Action State :=
  ActionPlus (expr_to_Prop if3)
  (Actions (perf_store (PMM_expr (expr_var i_var)) (expr_z 0%Z)) ;;
   Actions (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) ;;
   mem_init_loop3_spec) 
  (ActionID State).


Definition WF_mem_init_loop3' : WFProc expr_to_Prop Mc_Machine Psi mem_init_loop3_spec' mem_init_loop3.
unfold mem_init_loop3_spec'; unfold mem_init_loop3. auto 999.
Qed.


Definition mem_init_loop2_spec' : Action State :=
  ActionPlus (expr_to_Prop if2)
  (Actions (perf_store (PMM_expr (expr_var i_var)) (expr_z 1%Z)) ;;
   Actions (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) ;;
   mem_init_loop2_spec) 
  (ActionID State).


Definition WF_mem_init_loop2' : WFProc expr_to_Prop Mc_Machine Psi mem_init_loop2_spec' mem_init_loop2.
unfold mem_init_loop2_spec'; unfold mem_init_loop2. auto 999.
Qed.


Definition WF_mem_init : WFProc expr_to_Prop Mc_Machine Psi mem_sig_ald.PM_mem_init_spec mem_init.
apply WFProc_weaken with mem_init_spec'; [| apply WF_mem_init'].

unfold mem_sig_ald.PM_mem_init_spec, mem_init_spec', if1.
C_auf.
sact_structured.
sact_auf.

unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
split; intros; repeat (autorewrite with base c mem stack in *; simpl in *).

unfold StackCheck, StackGet; simpl.
repeat split; repeat autoclean; simpl in *; autosolve.

rewrite H3; simpl; auto.
rewrite H3; simpl; auto.
rewrite H3; simpl; auto.
omega.
rewrite H3; simpl; auto.
rewrite H6; simpl; auto.

rewrite H3; simpl; auto.
rewrite H6; simpl; auto.

rewrite H3; simpl; auto.
rewrite H6; simpl; auto.
unfold NPAGES; omega.

rewrite H3; simpl.
rewrite H6; simpl.
rewrite H9; simpl; auto.

(* rel *)
repeat split; repeat autoclean; simpl in *; autosolve.
unfold StackGet in *; simpl in *.
rewrite H7 in *; clear H7; simpl in *.
rewrite H10 in *; clear H10; simpl in *.
rewrite <- H3; autosolve.
apply H5.
apply ALE_State.PtDom_alloc; auto.
left.
set (xxx := ALE_State.PtDom_alloc l H13).
omega.

unfold StackGet in *; simpl in *.
rewrite H7 in *; clear H7; simpl in *.
rewrite H10 in *; clear H10; simpl in *.
rewrite <- H3; autosolve.
apply H5.
apply PE_State.PMMDom_alloc; auto.
left.
set (xxx := PE_State.PMMDom_alloc l H13).
omega.

unfold StackGet in *; simpl in *.
rewrite H7 in *; clear H7; simpl in *.
rewrite H10 in *; clear H10; simpl in *.
unfold setMem in *.
assert (ValidPg (Pg l) = true).
apply ValidPA_ValidPg; auto.
unfold ValidPg in H0.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
destruct (Zlt_dec (Pg l) 512%Z).
rewrite <- H3; autosolve.
destruct (Zlt_dec (Pg l) 100%Z).
rewrite <- H6; autosolve.
destruct (Zeq_dec (Pg l) 0%Z).
rewrite <- H9; autosolve.
rewrite e.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H8; autosolve.
rewrite H5; autosolve.
rewrite H2; autosolve.

unfold StackGet in *; simpl in *.
rewrite H7 in *; clear H7; simpl in *.
rewrite H10 in *; clear H10; simpl in *.
rewrite H4; simpl; auto.
Qed.

Opaque NPAGES.

Definition WF_mem_init_loop3 : WFProc expr_to_Prop Mc_Machine Psi mem_init_loop3_spec mem_init_loop3.
apply WFProc_weaken with mem_init_loop3_spec'; [| apply WF_mem_init_loop3'].

unfold mem_init_loop3_spec, mem_init_loop3_spec', if3.
C_auf.
sact_structured.
sact_auf.

split; simpl.
(* pre *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold SCheck_and, StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct, getVarCheck  in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
repeat (autorewrite with base c mem stack in *; simpl in *).
destruct H0.
left.
unfold setMemCheck, getVarCheck in *; simpl in *.
repeat split; autosolve; intros; repeat autoclean; simpl in *; try subst; try rewrite H3; simpl in *; autosolve.
apply PE_State.PMMDom_Valid; auto.

unfold ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
left; auto.
right; autosolve.
unfold ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

unfold SCheck_and, StackCheck, getVarCheck; simpl.
rewrite H3; autosolve.

Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold StackGet in *; simpl in *.
rewrite H3 in H2; autosolve.
unfold ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

right; subst.
red; intros.
red in H; simpl in *; repeat autoclean.
unfold SCheck_and, StackCheck, StackGet, getVarCheck in *; simpl in *.
rewrite H3 in *; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

(* rel *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold expr_to_Prop, SCheck_and, StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
unfold expr_to_Prop, SCheck_and, StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
unfold getVarCheck in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *.
repeat autoclean.
destruct H2.
unfold ValidPg in H.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
destruct H0; autosolve.
split; autosolve.
repeat autoclean; simpl in *.
repeat split; autosolve.
intros.
destruct (Zeq_dec z pg).
subst.
rewrite <- H2; autosolve.
unfold setMem.
rewrite Zeq_bool_refl; auto.

rewrite H5; autosolve.

apply H0; autosolve.
rewrite H5; autosolve.

intros.
rewrite <- H2; autosolve.
unfold setMem.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H5 in *; autosolve.

rewrite H7.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

clear H0.
subst.
rewrite H1 in *; simpl in *; autosolve; clear H1.
repeat split; autosolve.
intros; autosolve.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst; auto.
red; intros.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Definition WF_mem_init_loop2 : WFProc expr_to_Prop Mc_Machine Psi mem_init_loop2_spec mem_init_loop2.
apply WFProc_weaken with mem_init_loop2_spec'; [| apply WF_mem_init_loop2'].

unfold mem_init_loop2_spec, mem_init_loop2_spec', if2, expr_to_Prop.
C_auf.
sact_structured.
sact_auf.

split; simpl.
(* pre *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold setMemCheck, setMem; simpl in *.
destruct (Zeq_dec z 512%Z).

subst; right.
red; intros.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

left; repeat split; intros; repeat autoclean; autosolve; subst; simpl in *; autosolve.
apply PE_State.PMMDom_Valid; auto.
rewrite H3; autosolve.
rewrite H3; autosolve.
Transparent NPAGES.
unfold ValidPg in *.
unfold NPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

(* rel *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
destruct (Zeq_dec z 512%Z).

clear H0; subst.
rewrite H1; simpl in *; autosolve; clear H1 s'.
repeat split.
intros; autosolve.
repeat f_equal.
apply functional_extensionality; intros; unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.

clear H1; destruct H0; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; split; autosolve.
repeat autoclean; simpl in *.
rewrite H4; clear H4.
repeat split.
intros.
destruct (Zeq_dec z pg).
subst.
rewrite <- H1; autosolve.
unfold setMem.
rewrite Zeq_bool_refl; auto.
rewrite H5; autosolve.
apply H0; autosolve.
rewrite H5; autosolve.
intros.
rewrite <- H1; autosolve.
unfold setMem.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H5; autosolve.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.


Definition WF_mem_init_loop1 : WFProc expr_to_Prop Mc_Machine Psi mem_init_loop1_spec mem_init_loop1.
apply WFProc_weaken with mem_init_loop1_spec'; [| apply WF_mem_init_loop1'].

unfold mem_init_loop1_spec, mem_init_loop1_spec', if1, expr_to_Prop.
C_auf.
sact_structured.
sact_auf.

split; simpl.
(* pre *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold setMemCheck, setMem; simpl in *.
destruct (Zeq_dec z 100%Z).

subst; right.
red; intros.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

left; repeat split; intros; repeat autoclean; autosolve; subst; simpl in *; autosolve.
apply PE_State.PMMDom_Valid; auto.
rewrite H3; autosolve.
rewrite H3; autosolve.
Transparent NPAGES.
unfold ValidPg in *.
unfold NPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

(* rel *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
destruct (Zeq_dec z 100%Z).

clear H0; subst.
rewrite H1; simpl in *; autosolve; clear H1 s'.
repeat split.
intros; autosolve.
repeat f_equal.
apply functional_extensionality; intros; unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.

clear H1; destruct H0; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; split; autosolve.
repeat autoclean; simpl in *.
rewrite H4; clear H4.
repeat split.
intros.
destruct (Zeq_dec z pg).
subst.
rewrite <- H1; autosolve.
unfold setMem.
rewrite Zeq_bool_refl; auto.
rewrite H5; autosolve.
apply H0; autosolve.
rewrite H5; autosolve.
intros.
rewrite <- H1; autosolve.
unfold setMem.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H5; autosolve.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.











Definition mem_init_PSI : C_PM.SCAP.SpecHeap.SpecHeap := fun f =>
   match f with
   | mem_init_label => Some (mem_sig_ald.PM_mem_init_spec)
   | mem_init_loop1_label  => Some (mem_init_loop1_spec)
   | mem_init_loop2_label => Some (mem_init_loop2_spec)
   | mem_init_loop3_label => Some (mem_init_loop3_spec)
   | _ => None
   end.

Definition mem_init_CODE : C_PM.SCAP.CH.CodeHeap := fun f =>
   match f with
   | mem_init_label => Some (mem_init)
   | mem_init_loop1_label  => Some (mem_init_loop1)
   | mem_init_loop2_label => Some (mem_init_loop2)
   | mem_init_loop3_label => Some (mem_init_loop3)
   | _ => None
   end.

Hint Resolve ActionGTE_refl.

Lemma WF_mem_init_lib : 
    WFCode expr_to_Prop C_PM.Mc_Machine  
            C_PM.SCAP.SpecHeap.SpecHeap_empty
            mem_init_CODE
            mem_init_PSI.
set (x := C_PM.SCAP.SpecHeap.SpecHeap_empty_disjoint mem_init_PSI).
apply WFCode_1 with x.
rewrite C_PM.SCAP.SpecHeap.SpecHeap_join_empty.
intros.
red in H; simpl in H.
destruct f; simpl; autosolve.

inversion_clear H.
eexists; split; auto.
apply WFProc_strengthen with Psi.
red; intros.
destruct l; simpl; autosolve.
apply WF_mem_init.

inversion_clear H.
eexists; split; auto.
apply WFProc_strengthen with Psi.
red; intros.
destruct l; simpl; autosolve.
apply WF_mem_init_loop1.

inversion_clear H.
eexists; split; auto.
apply WFProc_strengthen with Psi.
red; intros.
destruct l; simpl; autosolve.
apply WF_mem_init_loop2.

inversion_clear H.
eexists; split; auto.
apply WFProc_strengthen with Psi.
red; intros.
destruct l; simpl; autosolve.
apply WF_mem_init_loop3.
Qed.

End mem_init.
