(* mem_lib_cert.v *)
(* Defines the certified module containing the memory allocator *)
(* M_{PE}, LIB_{PE} |- C^{mem} : SPEC^{mem} *)

Require Import label.
Require Import spec.

Require Import c.
Require Import pe.
Require Import pe_lib.
Require Import mach.mem_sig.

Require Import mem_alloc.
Require Import mem_free.

Require Import zbool_assist.
Require Import actsmp.

Definition mem_code : C_PE.SCAP.CodeHeap := fun l =>
  match l with
  | mem_free_label => Some mem_free.mem_free
  | mem_alloc_label => Some mem_alloc.mem_alloc
  | mem_alloc_loop_label => Some mem_alloc.mem_alloc_loop
  | mem_alloc_body_label => Some mem_alloc.mem_alloc_body
  | _ => None
  end.


Definition mem_spec : C_PE.SCAP.SpecHeap.SpecHeap := fun l =>
  match l with
  | mem_free_label => Some mem_sig.mem_free_spec
  | mem_alloc_label => Some mem_sig.mem_alloc_spec
  | mem_alloc_loop_label => Some mem_alloc.mem_alloc_loop_spec
  | mem_alloc_body_label => Some mem_alloc.mem_alloc_body_spec
  | _ => None
  end.

Lemma lib_mem_disj : C_PE.SCAP.SpecHeap.SpecHeap_disjoint PE_Lib.pe_lib_spec mem_spec.
red; intros.
destruct l; autosolve.
Qed.


Lemma WF_mem_lib : C_PE.SCAP.WFCode C_PE.State.expr_to_Prop 
      C_PE.Mc_Machine PE_Lib.pe_lib_spec mem_code mem_spec.
apply C_PE.SCAP.WFCode_1 with lib_mem_disj.
intros.
red in H.
destruct f; autosolve.

(* mem_alloc *)
inversion_clear H.
exists mem_alloc.mem_alloc.
split; auto.
apply C_PE.SCAP.Process.WFProc_strengthen with (mem_alloc.Psi).
red; intros.
destruct l; autosolve; simpl; try apply ActionGTE_refl.
apply mem_alloc.WF_mem_alloc.

(* mem_free *)
inversion_clear H.
exists mem_free.mem_free.
split; auto.
apply C_PE.SCAP.Process.WFProc_strengthen with (mem_free.Psi).
red; intros.
destruct l; autosolve; simpl; try apply ActionGTE_refl.
apply mem_free.WF_mem_free.

(* mem_alloc_loop *)
inversion_clear H.
exists mem_alloc.mem_alloc_loop.
split; auto.
apply C_PE.SCAP.Process.WFProc_strengthen with (mem_alloc.Psi).
red; intros.
destruct l; autosolve; simpl; try apply ActionGTE_refl.
apply mem_alloc.WF_mem_alloc_loop.

(* mem_alloc_body *)
inversion_clear H.
exists mem_alloc.mem_alloc_body.
split; auto.
apply C_PE.SCAP.Process.WFProc_strengthen with (mem_alloc.Psi).
red; intros.
destruct l; autosolve; simpl; try apply ActionGTE_refl.
apply mem_alloc.WF_mem_alloc_body.
Qed.
