(* pmap_init.v *)
(* Defines and certified the code of the pt_init procedure *)
(* Includes the code and certification of subprocedures - e.g. the loops of pt_init *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import spec.


Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import ald.
Require Import mach.mem_sig.

Module pmap_init.
Import C_ALD State Mem Stack Ops SCAP Process.

Require Import sact.
Require Import actsmp.


(***************************************************)


Definition i_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Import ald.
Import ale.
Import pe.

Opaque PTROOT PE_State.PTROOT ALE_State.PTROOT PE_State.PMM_data_loc.

Definition if1 := expr_nomem_binop LT (expr_nomem_var i_var) (expr_nomem_z NPAGES).
Definition if2 := expr_nomem_binop LT (expr_nomem_var i_var) (expr_nomem_z VPAGES).

Definition PT_expr (pg : expr) : expr :=
  (expr_binop PLUS (expr_z ALE_State.PTROOT) (expr_binop TIMES pg (expr_z 8%Z))).

Definition pmap_init_loop1_spec : Action State := 
 {| pre := (fun (s : State) => StackCheck (getVarCheck i_var) s = true /\
                                           ValidVPg (StackGet (getVar i_var) s) = true /\
                                           (forall l, ValidPA l = true ->
                                                        ALE_State.PtDom l = true -> s.(memory).(alloctbl) (Pg l)  = true) /\
                                           (StackGet (getVar i_var) s <= NPAGES)%Z); 
   rel := fun (s s' : State) => let i := StackGet (getVar i_var) s  in
                            (forall pg, (i <= pg < NPAGES)%Z -> (s'.(memory).(data) (PTROOT + pg* 8)%Z = pg%Z)) /\
                            (forall l, ((l < PTROOT + i  * 8)%Z \/ (l >= PTROOT + NPAGES*8)%Z) -> 
                                            s.(memory).(data) l = s'.(memory).(data) l) /\
                            (s.(memory).(alloctbl) = s'.(memory).(alloctbl)) /\
                            (s'.(stack) = setVar i_var NPAGES%Z s.(stack)) |}. 

Definition pmap_init_loop2_spec : Action State := 
 {| pre := (fun (s : State) => StackCheck (getVarCheck i_var) s = true /\ 
                                           (ValidVPg (StackGet (getVar i_var) s) = true \/ (StackGet (getVar i_var) s) = VPAGES) /\
                                           (forall l, ValidPA l = true -> 
                                                       ALE_State.PtDom l = true -> s.(memory).(alloctbl) (Pg l)  = true) /\
                                           (StackGet (getVar i_var) s <= VPAGES)%Z); 
   rel := fun (s s' : State) => let i := StackGet (getVar i_var) s  in
                            (forall pg, (i <= pg < VPAGES)%Z -> (s'.(memory).(data) (PTROOT + pg* 8)%Z = 0%Z)) /\
                            (forall l, ((l < PTROOT + i  * 8)%Z \/ (l >= PTROOT + VPAGES*8)%Z) -> 
                                            s.(memory).(data) l = s'.(memory).(data) l) /\
                            (s.(memory).(alloctbl) = s'.(memory).(alloctbl)) /\
                            (s'.(stack) = setVar i_var VPAGES%Z s.(stack)) |}. 

Require Import label.

Definition Psi (f : Label) : option (Action State) := 
   match f with
   | pmap_init_loop1_label  => Some (pmap_init_loop1_spec)
   | pmap_init_loop2_label => Some (pmap_init_loop2_spec)
   | _ => None
   end.


Definition pmap_init : Process :=
   Iperf (perf_readargs nil) :::
   Iperf (perf_assign i_var (expr_z 0%Z)) :::
   Icall pmap_init_loop1_label :::
   Icall pmap_init_loop2_label :::
   Iperf (perf_ret (expr_z 0%Z)).

Definition pmap_init_spec' : Action State :=
   Actions (perf_readargs nil) ;;
   Actions (perf_assign i_var (expr_z 0%Z)) ;;
   pmap_init_loop1_spec ;;
   pmap_init_loop2_spec ;;
   Actions (perf_ret (expr_z 0%Z)).

Definition WF_pmap_init' : WFProc expr_to_Prop Mc_Machine Psi pmap_init_spec' pmap_init.
unfold pmap_init_spec'; unfold pmap_init; auto 999.
Qed.

Require Import ald_lib.

Lemma PTROOT_domain : forall l, ALE_State.PtDom l = false -> (l < PTROOT \/ l >= PTROOT + VPAGES*8)%Z.
intros.
Transparent ALE_State.PTROOT common.PTROOT.
unfold ALE_State.PtDom, ALE_State.PTROOT, common.PTROOT, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.
Opaque ALE_State.PTROOT common.PTROOT.

Definition WF_pmap_init : WFProc expr_to_Prop Mc_Machine Psi ALD_Lib.ALD_pmap_init_spec pmap_init.
apply WFProc_weaken with pmap_init_spec'; [| apply WF_pmap_init'].

unfold ALD_Lib.ALD_pmap_init_spec, pmap_init_spec'.
C_auf.
sact_structured.
sact_auf.

unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.
split; intros; repeat (autorewrite with base c mem stack in *; simpl in *).

unfold StackCheck, StackGet; simpl.
repeat autoclean.
repeat split; intros; simpl in *; subst; simpl in *; autosolve.
repeat autoclean; simpl in *; autosolve.
repeat autoclean; simpl in *; autosolve.
repeat autoclean; simpl in *; autosolve.
apply H0; autosolve.
repeat autoclean; simpl in *; autosolve.
repeat destruct H1; simpl in *; subst; simpl in *.
decompose record H2; clear H2; subst; simpl in *.
rewrite H6 in *; simpl in *; auto.

repeat destruct H1; simpl in *; subst; simpl in *.
decompose record H2; clear H2; subst; simpl in *.
rewrite H6 in *; simpl in *; auto.

repeat destruct H1; simpl in *; subst; simpl in *.
decompose record H4; clear H4; subst; simpl in *.
rewrite <- H5.
apply H0; autosolve.

repeat destruct H1; simpl in *; subst; simpl in *.
decompose record H2; clear H2; subst; simpl in *.
rewrite H6; simpl in *.
unfold NPAGES, VPAGES; autosolve.

repeat destruct H1; simpl in *; subst; simpl in *.
decompose record H2; clear H2; subst; simpl in *.
decompose record H3; clear H3; subst; simpl in *.
rewrite H7; rewrite H10; simpl; auto.

(* rel *)
unfold StackGet in *; simpl in *.
repeat destruct H0; simpl in *; subst; simpl in *.
decompose record H2; clear H2.
decompose record H3; clear H3.
simpl in *.
destruct s; simpl in *.
destruct stack0; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
decompose record H; clear H.
rewrite H6 in *; rewrite H9 in *; simpl in *.
repeat split; autosolve.

intros.
unfold ValidPg in H.
unfold NPAGES in *.
rewrite <- H4; autosolve.
apply H2; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros.
apply H0.
unfold ValidVPg, ValidPg, NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros.
set (xxx := PTROOT_domain l H10).
unfold ValidVPg, ValidPg, NPAGES, VPAGES in *; simpl in *.
rewrite <- H4; autosolve.
rewrite <- H7; autosolve.

intros.
rewrite <- H1; auto.
rewrite <- H5; auto.
Qed.

Definition pmap_init_loop1 : Process :=
  Iplus (if1)
  (Iperf (perf_store (PT_expr (expr_var i_var)) (expr_var i_var%Z)) :::
   Iperf (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) :::
   Icall pmap_init_loop1_label) 
  Inil.

Definition pmap_init_loop1_spec' : Action State :=
  ActionPlus (expr_to_Prop if1)
  (Actions (perf_store (PT_expr (expr_var i_var)) (expr_var i_var%Z)) ;;
   Actions (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) ;;
   pmap_init_loop1_spec) 
  (ActionID State).


Definition pmap_init_loop2 : Process :=
  Iplus (if2)
  (Iperf (perf_store (PT_expr (expr_var i_var)) (expr_z 0%Z)) :::
   Iperf (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) :::
   Icall pmap_init_loop2_label) 
  Inil.

Definition pmap_init_loop2_spec' : Action State :=
  ActionPlus (expr_to_Prop if2)
  (Actions (perf_store (PT_expr (expr_var i_var)) (expr_z 0%Z)) ;;
   Actions (perf_assign i_var (expr_binop PLUS (expr_var i_var) (expr_z 1%Z))) ;;
   pmap_init_loop2_spec) 
  (ActionID State).

Definition WF_pmap_init_loop1' : WFProc expr_to_Prop Mc_Machine Psi pmap_init_loop1_spec' pmap_init_loop1.
unfold pmap_init_loop1_spec'; unfold pmap_init_loop1. auto 999.
Qed.

Definition WF_pmap_init_loop2' : WFProc expr_to_Prop Mc_Machine Psi pmap_init_loop2_spec' pmap_init_loop2.
unfold pmap_init_loop2_spec'; unfold pmap_init_loop2. auto 999.
Qed.

Definition WF_pmap_init_loop1 : WFProc expr_to_Prop Mc_Machine Psi pmap_init_loop1_spec pmap_init_loop1.
apply WFProc_weaken with pmap_init_loop1_spec'; [| apply WF_pmap_init_loop1'].

unfold pmap_init_loop1_spec, pmap_init_loop1_spec', if1, expr_to_Prop.
C_auf.
sact_structured.
sact_auf.

split; simpl.
(* pre *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold setMemCheck, ValidAddr, setMem; simpl in *.

destruct (Zeq_dec z NPAGES).
subst; right.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

left.
rewrite H1; autosolve.
repeat split; intros; repeat autoclean; auto; simpl in *; subst; simpl in *; autosolve.
apply ALE_State.PtDom_Valid; auto.
rewrite H4; auto.
rewrite H4; auto.
unfold ValidVPg in *.
unfold NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H1; auto.
rewrite H4; auto.
unfold NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply ALE_State.PtDom_Valid; auto.
apply ALE_State.PtDom_PTROOT; auto.

(* rel *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold setMemCheck, ValidAddr, setMem; simpl in *.

destruct (Zeq_dec z NPAGES).
subst.
clear H0; rewrite H1; simpl; autosolve.
repeat split; autosolve.
intros; autosolve.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.

clear H1.
destruct H0; autosolve.
repeat destruct H; simpl in *; subst; simpl in *.
rewrite H6 in *; simpl in *.
repeat destruct H0; simpl in *.
destruct H1.
repeat split; autosolve.

intros.
destruct (Zeq_dec z pg).
subst.
rewrite <- H0; autosolve.
rewrite Zeq_bool_refl; auto.
apply H; autosolve.

intros.
replace (ALE_State.PTROOT) with (PTROOT) in *; autosolve.
rewrite <- H0; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
unfold NPAGES in *; autosolve.

rewrite H7.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.
Qed.

Definition WF_pmap_init_loop2 : WFProc expr_to_Prop Mc_Machine Psi pmap_init_loop2_spec pmap_init_loop2.
apply WFProc_weaken with pmap_init_loop2_spec'; [| apply WF_pmap_init_loop2'].

unfold pmap_init_loop2_spec, pmap_init_loop2_spec', if2, expr_to_Prop.
C_auf.
sact_structured.
sact_auf.

split; simpl.
(* pre *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold setMemCheck, ValidAddr, setMem; simpl in *.

destruct (Zeq_dec z VPAGES).
subst; right.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

left.
rewrite H1; autosolve.
repeat split; intros; auto; try repeat destruct H; simpl in *; subst; simpl in *; autosolve.
apply ALE_State.PtDom_Valid; auto.
unfold ValidVPg in *.
unfold NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
destruct H0; autosolve.

rewrite H4; autosolve.
rewrite H4; autosolve.
destruct H0; autosolve.
unfold ValidVPg in *.
unfold NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
left; autosolve.
right; autosolve.

apply H1; auto.
rewrite H4; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

destruct H0; autosolve.
apply ALE_State.PtDom_Valid; autosolve.

destruct H0; autosolve.
apply ALE_State.PtDom_PTROOT; auto.

(* rel *)
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold StackCheck, StackGet, MemCheck, MemGet, StackAct, MemAct in *; simpl in *.
repeat autoclean.
destruct s; simpl in *.
destruct stack0; autosolve; simpl in *.
destruct f; autosolve; simpl in *.
makenew' (d i_var); autosolve; simpl in *; repeat autoclean.
intros; repeat (autorewrite with base c mem stack in *; simpl in *).
unfold setMemCheck, ValidAddr, setMem; simpl in *.

destruct (Zeq_dec z VPAGES).
subst.
clear H0; rewrite H1; simpl; autosolve.
repeat split; autosolve.
intros; autosolve.
repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.

clear H1.
destruct H0.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.
destruct H2; autosolve.
repeat destruct H; simpl in *; subst; simpl in *.
rewrite H6 in *; simpl in *.
decompose record H1; clear H1.
rewrite H8 in *; simpl in *; clear H8.
repeat split; autosolve.

intros.
destruct (Zeq_dec z pg).
subst.
rewrite <- H5; autosolve.
rewrite Zeq_bool_refl; auto.
apply H; autosolve.

intros.
rewrite <- H5; autosolve.
unfold NPAGES, VPAGES in *.
replace (ALE_State.PTROOT) with (PTROOT) in *; autosolve.
destruct H1; autosolve.
unfold NPAGES, VPAGES in *.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.

repeat f_equal.
apply functional_extensionality; intros.
unfold Var_beq.
Zeq_pull'; Zeq_norm; simpl in *; subst; autosolve.
Qed.


Definition pmap_init_PSI : C_ALD.SCAP.SpecHeap.SpecHeap := fun f =>
   match f with
   | pmap_init_label => Some (ALD_Lib.ALD_pmap_init_spec)
   | pmap_init_loop1_label  => Some (pmap_init_loop1_spec)
   | pmap_init_loop2_label => Some (pmap_init_loop2_spec)
   | _ => None
   end.

Definition pmap_init_CODE : C_ALD.SCAP.CH.CodeHeap := fun f =>
   match f with
   | pmap_init_label => Some (pmap_init)
   | pmap_init_loop1_label  => Some (pmap_init_loop1)
   | pmap_init_loop2_label => Some (pmap_init_loop2)
   | _ => None
   end.

Hint Resolve ActionGTE_refl.

Lemma WF_pmap_init_lib : 
    WFCode expr_to_Prop C_ALD.Mc_Machine  
            C_ALD.SCAP.SpecHeap.SpecHeap_empty
            pmap_init_CODE
            pmap_init_PSI.
set (x := C_ALD.SCAP.SpecHeap.SpecHeap_empty_disjoint pmap_init_PSI).
apply WFCode_1 with x.
rewrite C_ALD.SCAP.SpecHeap.SpecHeap_join_empty.
intros.
red in H; simpl in H.
destruct f; simpl; autosolve.

inversion_clear H.
eexists; split; auto.
apply WFProc_strengthen with Psi.
red; intros.
destruct l; simpl; autosolve.
apply WF_pmap_init.

inversion_clear H.
eexists; split; auto.
apply WFProc_strengthen with Psi.
red; intros.
destruct l; simpl; autosolve.
apply WF_pmap_init_loop1.

inversion_clear H.
eexists; split; auto.
apply WFProc_strengthen with Psi.
red; intros.
destruct l; simpl; autosolve.
apply WF_pmap_init_loop2.
Qed.

End pmap_init.
