(* pmap_lib_cert.v *)
(* Defines the certified module containing the page table driver *)
(* Mach_{ALE}, LIB_{ALE} |- C^{pt} : SPEC^{pt} *)

Require Import label.
Require Import c.
Require Import ale.
Require Import ale_lib.
Require Import mach.pmap_sig.

Require Import pmap_set.
Require Import pmap_lookup.

Require Import zbool_assist.
Require Import actsmp.

Definition pmap_lib_code : C_ALE.SCAP.CodeHeap := fun l =>
  match l with
  | pmap_set_label => Some pmap_set.pmap_set
  | pmap_lookup_label => Some pmap_lookup.pmap_lookup
  | _ => None
  end.

Lemma WF_pmap_lib : C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop 
                       C_ALE.Mc_Machine ALE_Lib.ale_lib_spec pmap_lib_code pmap_sig.pmap_spec.
apply C_ALE.SCAP.WFCode_1 with pmap_sig.lib_pmap_disj.
intros.
red in H.
destruct f; autosolve.

(* pmap_set *)
inversion_clear H.
exists pmap_set.pmap_set.
split; auto.
apply C_ALE.SCAP.Process.WFProc_strengthen with ALE_Lib.ale_lib_spec.
apply C_ALE.SCAP.SpecHeap.SpecHeap_stronger_add.
apply C_ALE.SCAP.Process.WFProc_strengthen with (pmap_set.Psi).
red; intros.
destruct l; autosolve.
apply pmap_set.WF_pmap_set.

(* pmap_lookup *)
inversion_clear H.
exists pmap_lookup.pmap_lookup.
split; auto.
apply C_ALE.SCAP.Process.WFProc_strengthen with ALE_Lib.ale_lib_spec.
apply C_ALE.SCAP.SpecHeap.SpecHeap_stronger_add.
apply C_ALE.SCAP.Process.WFProc_strengthen with (pmap_set.Psi).
red; intros.
destruct l; autosolve.
apply pmap_lookup.WF_pmap_lookup.
Qed.
