(* pmap_lookup.v *)
(* Defines and certifies the pt_lookup procedure *)

Require Import List.
Require Import Bool.
Require Import NArith.
Require Import ZArith.
Require Import IfProp.

Require Import zbool_assist.
Require Import base_simp.
Require Import Omega.

Require Import spec.

Require Import Logic.FunctionalExtensionality.

Require Import c.
Require Import ale.
Require Import mach.pmap_sig.

Module pmap_lookup.
Import C_ALE State Mem Stack Ops SCAP Process.

Require Import sact.
Require Import actsmp.

Opaque PTROOT ALE_State.PTROOT.
(***************************************************)

Definition Psi (f : Label) : option (Action State) := 
   match f with
   | _ => None
   end.

Definition junk_var : Var := 1%Z.
Definition page_var : Var := 2%Z.
Definition entry_var : Var := 3%Z.

Notation "a ::: b" := (Iseq a b) (at level 80).
Notation "a ;; b" := (ActionChain a b) (at level 80).

Definition ch := (fun s' => (StackCheck (getVarCheck entry_var) s') && (Zeq_bool (StackGet (getVar entry_var) s') 0%Z)).

Definition pmap_lookup : Process :=
  Iperf (perf_readargs (page_var :: nil)) :::
  Iperf (perf_assign entry_var 
                 (expr_mem 
                      (expr_binop PLUS (expr_z PTROOT) 
                               (expr_binop TIMES (expr_var page_var) (expr_z 8%Z))
                       )
                 )
           ) :::
  Iperf (perf_ret (expr_var entry_var)).

Definition pmap_lookup_spec' : Action State :=
  Actions (perf_readargs (page_var :: nil)) ;;
  Actions (perf_assign entry_var 
                 (expr_mem 
                      (expr_binop PLUS (expr_z PTROOT) 
                               (expr_binop TIMES (expr_var page_var) (expr_z 8%Z))
                       )
                 )
           ) ;;
  Actions (perf_ret (expr_var entry_var)).

Definition WF_pmap_lookup' : WFProc expr_to_Prop Mc_Machine Psi pmap_lookup_spec' pmap_lookup.
unfold pmap_lookup_spec'; unfold pmap_lookup; auto 999.
Qed.

Definition WF_pmap_lookup : WFProc expr_to_Prop Mc_Machine Psi pmap_sig.pmap_lookup_spec pmap_lookup.
apply WFProc_weaken with pmap_lookup_spec'; [|apply WF_pmap_lookup'].
unfold pmap_sig.pmap_lookup_spec, pmap_lookup_spec', ch.

C_auf.
sact_structured.
sact_auf.
unfold StackAct, MemAct, MemCheck, StackCheck, StackGet, MemGet in *; simpl in *.

split; intros; simpl in *; repeat (autorewrite with c stack mem in *; simpl in *).

repeat autoclean.
repeat split; intros; subst; simpl in *; autosolve.
unfold ALE_State.getMemCheck.
unfold vaCheck, transCheck, paCheck, isPhys.
assert (ValidPA (PTROOT + nth 0 (getArgs (stack s)) 0%Z * 8) = true).
apply PtDom_Valid; auto.
assert (ValidVA (PTROOT + nth 0 (getArgs (stack s)) 0%Z * 8) = true).
apply ValidPA_ValidVA; auto.
rewrite H2; auto.
rewrite H3; try apply PtDom_PTROOT; auto.
unfold getTLB in H4; rewrite H4; simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite ValidPA_ValidPg; autosolve.
rewrite H6; simpl.
rewrite H5; simpl; auto.
repeat autoclean; simpl in *; autosolve.

(* rel *)
repeat autoclean; simpl in *.
f_equal.
Qed.

End pmap_lookup.
